#!/usr/bin/env python
# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Defines functionality for pipelined execution of interfaces

The `Workflow` class provides core functionality for batch processing.
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import str, bytes, open

import os
import os.path as op
import sys
from datetime import datetime
from copy import deepcopy
import pickle
import shutil

import numpy as np

from ... import config, logging
from ...utils.misc import str2bool
from ...utils.functions import (getsource, create_function_from_source)

from ...interfaces.base import (traits, TraitedSpec, TraitDictObject,
                                TraitListObject)
from ...utils.filemanip import save_json, makedirs, to_str
from .utils import (generate_expanded_graph, export_graph, write_workflow_prov,
                    write_workflow_resources, format_dot, topological_sort,
                    get_print_name, merge_dict, format_node)

from .base import EngineBase
from .nodes import MapNode

# Py2 compat: http://python-future.org/compatible_idioms.html#collections-counter-and-ordereddict
from future import standard_library
standard_library.install_aliases()

logger = logging.getLogger('nipype.workflow')


class Workflow(EngineBase):
    """Controls the setup and execution of a pipeline of processes."""

    def __init__(self, name, base_dir=None):
        """Create a workflow object.

        Parameters
        ----------
        name : alphanumeric string
            unique identifier for the workflow
        base_dir : string, optional
            path to workflow storage

        """
        import networkx as nx
        super(Workflow, self).__init__(name, base_dir)
        self._graph = nx.DiGraph()

    # PUBLIC API
    def clone(self, name):
        """Clone a workflow

        .. note::

          Will reset attributes used for executing workflow. See
          _init_runtime_fields.

        Parameters
        ----------

        name: alphanumeric name
            unique name for the workflow

        """
        clone = super(Workflow, self).clone(name)
        clone._reset_hierarchy()
        return clone

    # Graph creation functions
    def connect(self, *args, **kwargs):
        """Connect nodes in the pipeline.

        This routine also checks if inputs and outputs are actually provided by
        the nodes that are being connected.

        Creates edges in the directed graph using the nodes and edges specified
        in the `connection_list`.  Uses the NetworkX method
        DiGraph.add_edges_from.

        Parameters
        ----------

        args : list or a set of four positional arguments

            Four positional arguments of the form::

              connect(source, sourceoutput, dest, destinput)

            source : nodewrapper node
            sourceoutput : string (must be in source.outputs)
            dest : nodewrapper node
            destinput : string (must be in dest.inputs)

            A list of 3-tuples of the following form::

             [(source, target,
                 [('sourceoutput/attribute', 'targetinput'),
                 ...]),
             ...]

            Or::

             [(source, target, [(('sourceoutput1', func, arg2, ...),
                                         'targetinput'), ...]),
             ...]
             sourceoutput1 will always be the first argument to func
             and func will be evaluated and the results sent ot targetinput

             currently func needs to define all its needed imports within the
             function as we use the inspect module to get at the source code
             and execute it remotely
        """
        if len(args) == 1:
            connection_list = args[0]
        elif len(args) == 4:
            connection_list = [(args[0], args[2], [(args[1], args[3])])]
        else:
            raise TypeError('connect() takes either 4 arguments, or 1 list of'
                            ' connection tuples (%d args given)' % len(args))

        disconnect = False
        if kwargs:
            disconnect = kwargs.get('disconnect', False)

        if disconnect:
            self.disconnect(connection_list)
            return

        newnodes = []
        for srcnode, destnode, _ in connection_list:
            if self in [srcnode, destnode]:
                msg = ('Workflow connect cannot contain itself as node:'
                       ' src[%s] dest[%s] workflow[%s]') % (srcnode, destnode,
                                                            self.name)

                raise IOError(msg)
            if (srcnode not in newnodes) and not self._has_node(srcnode):
                newnodes.append(srcnode)
            if (destnode not in newnodes) and not self._has_node(destnode):
                newnodes.append(destnode)
        if newnodes:
            self._check_nodes(newnodes)
            for node in newnodes:
                if node._hierarchy is None:
                    node._hierarchy = self.name
        not_found = []
        connected_ports = {}
        for srcnode, destnode, connects in connection_list:
            if destnode not in connected_ports:
                connected_ports[destnode] = []
            # check to see which ports of destnode are already
            # connected.
            if not disconnect and (destnode in self._graph.nodes()):
                for edge in self._graph.in_edges(destnode):
                    data = self._graph.get_edge_data(*edge)
                    for sourceinfo, destname in data['connect']:
                        if destname not in connected_ports[destnode]:
                            connected_ports[destnode] += [destname]
            for source, dest in connects:
                # Currently datasource/sink/grabber.io modules
                # determine their inputs/outputs depending on
                # connection settings.  Skip these modules in the check
                if dest in connected_ports[destnode]:
                    raise Exception("""\
Trying to connect %s:%s to %s:%s but input '%s' of node '%s' is already
connected.
""" % (srcnode, source, destnode, dest, dest, destnode))
                if not (hasattr(destnode, '_interface') and
                        ('.io' in str(destnode._interface.__class__) or any([
                            '.io' in str(val)
                            for val in destnode._interface.__class__.__bases__
                        ]))):
                    if not destnode._check_inputs(dest):
                        not_found.append(['in', destnode.name, dest])
                if not (hasattr(srcnode, '_interface') and
                        ('.io' in str(srcnode._interface.__class__) or any([
                            '.io' in str(val)
                            for val in srcnode._interface.__class__.__bases__
                        ]))):
                    if isinstance(source, tuple):
                        # handles the case that source is specified
                        # with a function
                        sourcename = source[0]
                    elif isinstance(source, (str, bytes)):
                        sourcename = source
                    else:
                        raise Exception(
                            ('Unknown source specification in '
                             'connection from output of %s') % srcnode.name)
                    if sourcename and not srcnode._check_outputs(sourcename):
                        not_found.append(['out', srcnode.name, sourcename])
                connected_ports[destnode] += [dest]
        infostr = []
        for info in not_found:
            infostr += [
                "Module %s has no %sput called %s\n" % (info[1], info[0],
                                                        info[2])
            ]
        if not_found:
            raise Exception(
                '\n'.join(['Some connections were not found'] + infostr))

        # turn functions into strings
        for srcnode, destnode, connects in connection_list:
            for idx, (src, dest) in enumerate(connects):
                if isinstance(src,
                              tuple) and not isinstance(src[1], (str, bytes)):
                    function_source = getsource(src[1])
                    connects[idx] = ((src[0], function_source, src[2:]), dest)

        # add connections
        for srcnode, destnode, connects in connection_list:
            edge_data = self._graph.get_edge_data(srcnode, destnode, None)
            if edge_data:
                logger.debug('(%s, %s): Edge data exists: %s', srcnode,
                             destnode, to_str(edge_data))
                for data in connects:
                    if data not in edge_data['connect']:
                        edge_data['connect'].append(data)
                    if disconnect:
                        logger.debug('Removing connection: %s', to_str(data))
                        edge_data['connect'].remove(data)
                if edge_data['connect']:
                    self._graph.add_edges_from([(srcnode, destnode,
                                                 edge_data)])
                else:
                    # pass
                    logger.debug('Removing connection: %s->%s', srcnode,
                                 destnode)
                    self._graph.remove_edges_from([(srcnode, destnode)])
            elif not disconnect:
                logger.debug('(%s, %s): No edge data', srcnode, destnode)
                self._graph.add_edges_from([(srcnode, destnode, {
                    'connect': connects
                })])
            edge_data = self._graph.get_edge_data(srcnode, destnode)
            logger.debug('(%s, %s): new edge data: %s', srcnode, destnode,
                         to_str(edge_data))

    def disconnect(self, *args):
        """Disconnect nodes
        See the docstring for connect for format.
        """
        if len(args) == 1:
            connection_list = args[0]
        elif len(args) == 4:
            connection_list = [(args[0], args[2], [(args[1], args[3])])]
        else:
            raise TypeError('disconnect() takes either 4 arguments, or 1 list '
                            'of connection tuples (%d args given)' % len(args))

        for srcnode, dstnode, conn in connection_list:
            logger.debug('disconnect(): %s->%s %s', srcnode, dstnode,
                         to_str(conn))
            if self in [srcnode, dstnode]:
                raise IOError(
                    'Workflow connect cannot contain itself as node: src[%s] '
                    'dest[%s] workflow[%s]') % (srcnode, dstnode, self.name)

            # If node is not in the graph, not connected
            if not self._has_node(srcnode) or not self._has_node(dstnode):
                continue

            edge_data = self._graph.get_edge_data(srcnode, dstnode, {
                'connect': []
            })
            ed_conns = [(c[0], c[1]) for c in edge_data['connect']]

            remove = []
            for edge in conn:
                if edge in ed_conns:
                    # idx = ed_conns.index(edge)
                    remove.append((edge[0], edge[1]))

            logger.debug('disconnect(): remove list %s', to_str(remove))
            for el in remove:
                edge_data['connect'].remove(el)
                logger.debug('disconnect(): removed connection %s', to_str(el))

            if not edge_data['connect']:
                self._graph.remove_edge(srcnode, dstnode)
            else:
                self._graph.add_edges_from([(srcnode, dstnode, edge_data)])

    def add_nodes(self, nodes):
        """ Add nodes to a workflow

        Parameters
        ----------
        nodes : list
            A list of EngineBase-based objects
        """
        newnodes = []
        all_nodes = self._get_all_nodes()
        for node in nodes:
            if self._has_node(node):
                raise IOError('Node %s already exists in the workflow' % node)
            if isinstance(node, Workflow):
                for subnode in node._get_all_nodes():
                    if subnode in all_nodes:
                        raise IOError(('Subnode %s of node %s already exists '
                                       'in the workflow') % (subnode, node))
            newnodes.append(node)
        if not newnodes:
            logger.debug('no new nodes to add')
            return
        for node in newnodes:
            if not issubclass(node.__class__, EngineBase):
                raise Exception('Node %s must be a subclass of EngineBase',
                                node)
        self._check_nodes(newnodes)
        for node in newnodes:
            if node._hierarchy is None:
                node._hierarchy = self.name
        self._graph.add_nodes_from(newnodes)

    def remove_nodes(self, nodes):
        """ Remove nodes from a workflow

        Parameters
        ----------
        nodes : list
            A list of EngineBase-based objects
        """
        self._graph.remove_nodes_from(nodes)

    # Input-Output access
    @property
    def inputs(self):
        return self._get_inputs()

    @property
    def outputs(self):
        return self._get_outputs()

    def get_node(self, name):
        """Return an internal node by name
        """
        nodenames = name.split('.')
        nodename = nodenames[0]
        outnode = [
            node for node in self._graph.nodes()
            if str(node).endswith('.' + nodename)
        ]
        if outnode:
            outnode = outnode[0]
            if nodenames[1:] and issubclass(outnode.__class__, Workflow):
                outnode = outnode.get_node('.'.join(nodenames[1:]))
        else:
            outnode = None
        return outnode

    def list_node_names(self):
        """List names of all nodes in a workflow
        """
        import networkx as nx
        outlist = []
        for node in nx.topological_sort(self._graph):
            if isinstance(node, Workflow):
                outlist.extend([
                    '.'.join((node.name, nodename))
                    for nodename in node.list_node_names()
                ])
            else:
                outlist.append(node.name)
        return sorted(outlist)

    def write_graph(self,
                    dotfilename='graph.dot',
                    graph2use='hierarchical',
                    format="png",
                    simple_form=True):
        """Generates a graphviz dot file and a png file

        Parameters
        ----------

        graph2use: 'orig', 'hierarchical' (default), 'flat', 'exec', 'colored'
            orig - creates a top level graph without expanding internal
            workflow nodes;
            flat - expands workflow nodes recursively;
            hierarchical - expands workflow nodes recursively with a
            notion on hierarchy;
            colored - expands workflow nodes recursively with a
            notion on hierarchy in color;
            exec - expands workflows to depict iterables

        format: 'png', 'svg'

        simple_form: boolean (default: True)
            Determines if the node name used in the graph should be of the form
            'nodename (package)' when True or 'nodename.Class.package' when
            False.

        """
        graphtypes = ['orig', 'flat', 'hierarchical', 'exec', 'colored']
        if graph2use not in graphtypes:
            raise ValueError('Unknown graph2use keyword. Must be one of: ' +
                             str(graphtypes))
        base_dir, dotfilename = op.split(dotfilename)
        if base_dir == '':
            if self.base_dir:
                base_dir = self.base_dir
                if self.name:
                    base_dir = op.join(base_dir, self.name)
            else:
                base_dir = os.getcwd()
        base_dir = makedirs(base_dir, exist_ok=True)
        if graph2use in ['hierarchical', 'colored']:
            if self.name[:1].isdigit():  # these graphs break if int
                raise ValueError('{} graph failed, workflow name cannot begin '
                                 'with a number'.format(graph2use))
            dotfilename = op.join(base_dir, dotfilename)
            self.write_hierarchical_dotfile(
                dotfilename=dotfilename,
                colored=graph2use == "colored",
                simple_form=simple_form)
            outfname = format_dot(dotfilename, format=format)
        else:
            graph = self._graph
            if graph2use in ['flat', 'exec']:
                graph = self._create_flat_graph()
            if graph2use == 'exec':
                graph = generate_expanded_graph(deepcopy(graph))
            outfname = export_graph(
                graph,
                base_dir,
                dotfilename=dotfilename,
                format=format,
                simple_form=simple_form)

        logger.info(
            'Generated workflow graph: %s (graph2use=%s, simple_form=%s).' %
            (outfname, graph2use, simple_form))
        return outfname

    def write_hierarchical_dotfile(self,
                                   dotfilename=None,
                                   colored=False,
                                   simple_form=True):
        dotlist = ['digraph %s{' % self.name]
        dotlist.append(
            self._get_dot(
                prefix='  ', colored=colored, simple_form=simple_form))
        dotlist.append('}')
        dotstr = '\n'.join(dotlist)
        if dotfilename:
            fp = open(dotfilename, 'wt')
            fp.writelines(dotstr)
            fp.close()
        else:
            logger.info(dotstr)

    def export(self,
               filename=None,
               prefix="output",
               format="python",
               include_config=False):
        """Export object into a different format

        Parameters
        ----------
        filename: string
           file to save the code to; overrides prefix
        prefix: string
           prefix to use for output file
        format: string
           one of "python"
        include_config: boolean
           whether to include node and workflow config values

        """
        import networkx as nx
        formats = ["python"]
        if format not in formats:
            raise ValueError('format must be one of: %s' % '|'.join(formats))
        flatgraph = self._create_flat_graph()
        nodes = nx.topological_sort(flatgraph)

        all_lines = None
        lines = ['# Workflow']
        importlines = [
            'from nipype.pipeline.engine import Workflow, '
            'Node, MapNode'
        ]
        functions = {}
        if format == "python":
            connect_template = '%s.connect(%%s, %%s, %%s, "%%s")' % self.name
            connect_template2 = '%s.connect(%%s, "%%s", %%s, "%%s")' \
                                % self.name
            wfdef = '%s = Workflow("%s")' % (self.name, self.name)
            lines.append(wfdef)
            if include_config:
                lines.append('%s.config = %s' % (self.name, self.config))
            for idx, node in enumerate(nodes):
                nodename = node.fullname.replace('.', '_')
                # write nodes
                nodelines = format_node(
                    node, format='python', include_config=include_config)
                for line in nodelines:
                    if line.startswith('from'):
                        if line not in importlines:
                            importlines.append(line)
                    else:
                        lines.append(line)
                # write connections
                for u, _, d in flatgraph.in_edges(nbunch=node, data=True):
                    for cd in d['connect']:
                        if isinstance(cd[0], tuple):
                            args = list(cd[0])
                            if args[1] in functions:
                                funcname = functions[args[1]]
                            else:
                                func = create_function_from_source(args[1])
                                funcname = [
                                    name for name in func.__globals__
                                    if name != '__builtins__'
                                ][0]
                                functions[args[1]] = funcname
                            args[1] = funcname
                            args = tuple([arg for arg in args if arg])
                            line_args = (u.fullname.replace('.', '_'), args,
                                         nodename, cd[1])
                            line = connect_template % line_args
                            line = line.replace("'%s'" % funcname, funcname)
                            lines.append(line)
                        else:
                            line_args = (u.fullname.replace('.', '_'), cd[0],
                                         nodename, cd[1])
                            lines.append(connect_template2 % line_args)
            functionlines = ['# Functions']
            for function in functions:
                functionlines.append(pickle.loads(function).rstrip())
            all_lines = importlines + functionlines + lines

            if not filename:
                filename = '%s%s.py' % (prefix, self.name)
            with open(filename, 'wt') as fp:
                fp.writelines('\n'.join(all_lines))
        return all_lines

    def run(self, plugin=None, plugin_args=None, updatehash=False):
        """ Execute the workflow

        Parameters
        ----------

        plugin: plugin name or object
            Plugin to use for execution. You can create your own plugins for
            execution.
        plugin_args : dictionary containing arguments to be sent to plugin
            constructor. see individual plugin doc strings for details.
        """
        if plugin is None:
            plugin = config.get('execution', 'plugin')
        if not isinstance(plugin, (str, bytes)):
            runner = plugin
            plugin = runner.__class__.__name__[:-len('Plugin')]
            plugin_args = runner.plugin_args
        else:
            name = '.'.join(__name__.split('.')[:-2] + ['plugins'])
            try:
                __import__(name)
            except ImportError:
                msg = 'Could not import plugin module: %s' % name
                logger.error(msg)
                raise ImportError(msg)
            else:
                plugin_mod = getattr(sys.modules[name], '%sPlugin' % plugin)
                runner = plugin_mod(plugin_args=plugin_args)
        flatgraph = self._create_flat_graph()
        self.config = merge_dict(deepcopy(config._sections), self.config)
        logger.info('Workflow %s settings: %s', self.name,
                    to_str(sorted(self.config)))
        self._set_needed_outputs(flatgraph)
        execgraph = generate_expanded_graph(deepcopy(flatgraph))
        for index, node in enumerate(execgraph.nodes()):
            node.config = merge_dict(deepcopy(self.config), node.config)
            node.base_dir = self.base_dir
            node.index = index
            if isinstance(node, MapNode):
                node.use_plugin = (plugin, plugin_args)
        self._configure_exec_nodes(execgraph)
        if str2bool(self.config['execution']['create_report']):
            self._write_report_info(self.base_dir, self.name, execgraph)
        runner.run(execgraph, updatehash=updatehash, config=self.config)
        datestr = datetime.utcnow().strftime('%Y%m%dT%H%M%S')
        if str2bool(self.config['execution']['write_provenance']):
            prov_base = op.join(self.base_dir,
                                'workflow_provenance_%s' % datestr)
            logger.info('Provenance file prefix: %s' % prov_base)
            write_workflow_prov(execgraph, prov_base, format='all')

        if config.resource_monitor:
            base_dir = self.base_dir or os.getcwd()
            write_workflow_resources(
                execgraph,
                filename=op.join(base_dir, self.name, 'resource_monitor.json'))
        return execgraph

    # PRIVATE API AND FUNCTIONS

    def _write_report_info(self, workingdir, name, graph):
        if workingdir is None:
            workingdir = os.getcwd()
        report_dir = op.join(workingdir, name)
        makedirs(report_dir, exist_ok=True)
        shutil.copyfile(
            op.join(op.dirname(__file__), 'report_template.html'),
            op.join(report_dir, 'index.html'))
        shutil.copyfile(
            op.join(op.dirname(__file__), '..', '..', 'external', 'd3.js'),
            op.join(report_dir, 'd3.js'))
        nodes, groups = topological_sort(graph, depth_first=True)
        graph_file = op.join(report_dir, 'graph1.json')
        json_dict = {'nodes': [], 'links': [], 'groups': [], 'maxN': 0}
        for i, node in enumerate(nodes):
            report_file = "%s/_report/report.rst" % \
                          node.output_dir().replace(report_dir, '')
            result_file = "%s/result_%s.pklz" % \
                          (node.output_dir().replace(report_dir, ''),
                           node.name)
            json_dict['nodes'].append(
                dict(
                    name='%d_%s' % (i, node.name),
                    report=report_file,
                    result=result_file,
                    group=groups[i]))
        maxN = 0
        for gid in np.unique(groups):
            procs = [i for i, val in enumerate(groups) if val == gid]
            N = len(procs)
            if N > maxN:
                maxN = N
            json_dict['groups'].append(
                dict(procs=procs, total=N, name='Group_%05d' % gid))
        json_dict['maxN'] = maxN
        for u, v in graph.in_edges():
            json_dict['links'].append(
                dict(source=nodes.index(u), target=nodes.index(v), value=1))
        save_json(graph_file, json_dict)
        graph_file = op.join(report_dir, 'graph.json')
        # Avoid RuntimeWarning: divide by zero encountered in log10
        num_nodes = len(nodes)
        if num_nodes > 0:
            index_name = np.ceil(np.log10(num_nodes)).astype(int)
        else:
            index_name = 0
        template = '%%0%dd_' % index_name

        def getname(u, i):
            name_parts = u.fullname.split('.')
            # return '.'.join(name_parts[:-1] + [template % i + name_parts[-1]])
            return template % i + name_parts[-1]

        json_dict = []
        for i, node in enumerate(nodes):
            imports = []
            for u, v in graph.in_edges(nbunch=node):
                imports.append(getname(u, nodes.index(u)))
            json_dict.append(
                dict(
                    name=getname(node, i),
                    size=1,
                    group=groups[i],
                    imports=imports))
        save_json(graph_file, json_dict)

    def _set_needed_outputs(self, graph):
        """Initialize node with list of which outputs are needed."""
        rm_outputs = self.config['execution']['remove_unnecessary_outputs']
        if not str2bool(rm_outputs):
            return
        for node in graph.nodes():
            node.needed_outputs = []
            for edge in graph.out_edges(node):
                data = graph.get_edge_data(*edge)
                sourceinfo = [
                    v1[0] if isinstance(v1, tuple) else v1
                    for v1, v2 in data['connect']
                ]
                node.needed_outputs += [
                    v for v in sourceinfo if v not in node.needed_outputs
                ]
            if node.needed_outputs:
                node.needed_outputs = sorted(node.needed_outputs)

    def _configure_exec_nodes(self, graph):
        """Ensure that each node knows where to get inputs from
        """
        for node in graph.nodes():
            node.input_source = {}
            for edge in graph.in_edges(node):
                data = graph.get_edge_data(*edge)
                for sourceinfo, field in data['connect']:
                    node.input_source[field] = \
                        (op.join(edge[0].output_dir(),
                                 'result_%s.pklz' % edge[0].name),
                         sourceinfo)

    def _check_nodes(self, nodes):
        """Checks if any of the nodes are already in the graph

        """
        node_names = [node.name for node in self._graph.nodes()]
        node_lineage = [node._hierarchy for node in self._graph.nodes()]
        for node in nodes:
            if node.name in node_names:
                idx = node_names.index(node.name)
                try:
                    this_node_lineage = node_lineage[idx]
                except IndexError:
                    raise IOError(
                        'Duplicate node name "%s" found.' % node.name)
                else:
                    if this_node_lineage in [node._hierarchy, self.name]:
                        raise IOError(
                            'Duplicate node name "%s" found.' % node.name)
            else:
                node_names.append(node.name)

    def _has_attr(self, parameter, subtype='in'):
        """Checks if a parameter is available as an input or output
        """
        if subtype == 'in':
            subobject = self.inputs
        else:
            subobject = self.outputs
        attrlist = parameter.split('.')
        cur_out = subobject
        for attr in attrlist:
            if not hasattr(cur_out, attr):
                return False
            cur_out = getattr(cur_out, attr)
        return True

    def _get_parameter_node(self, parameter, subtype='in'):
        """Returns the underlying node corresponding to an input or
        output parameter
        """
        if subtype == 'in':
            subobject = self.inputs
        else:
            subobject = self.outputs
        attrlist = parameter.split('.')
        cur_out = subobject
        for attr in attrlist[:-1]:
            cur_out = getattr(cur_out, attr)
        return cur_out.traits()[attrlist[-1]].node

    def _check_outputs(self, parameter):
        return self._has_attr(parameter, subtype='out')

    def _check_inputs(self, parameter):
        return self._has_attr(parameter, subtype='in')

    def _get_inputs(self):
        """Returns the inputs of a workflow

        This function does not return any input ports that are already
        connected
        """
        inputdict = TraitedSpec()
        for node in self._graph.nodes():
            inputdict.add_trait(node.name, traits.Instance(TraitedSpec))
            if isinstance(node, Workflow):
                setattr(inputdict, node.name, node.inputs)
            else:
                taken_inputs = []
                for _, _, d in self._graph.in_edges(nbunch=node, data=True):
                    for cd in d['connect']:
                        taken_inputs.append(cd[1])
                unconnectedinputs = TraitedSpec()
                for key, trait in list(node.inputs.items()):
                    if key not in taken_inputs:
                        unconnectedinputs.add_trait(key,
                                                    traits.Trait(
                                                        trait, node=node))
                        value = getattr(node.inputs, key)
                        setattr(unconnectedinputs, key, value)
                setattr(inputdict, node.name, unconnectedinputs)
                getattr(inputdict, node.name).on_trait_change(self._set_input)
        return inputdict

    def _get_outputs(self):
        """Returns all possible output ports that are not already connected
        """
        outputdict = TraitedSpec()
        for node in self._graph.nodes():
            outputdict.add_trait(node.name, traits.Instance(TraitedSpec))
            if isinstance(node, Workflow):
                setattr(outputdict, node.name, node.outputs)
            elif node.outputs:
                outputs = TraitedSpec()
                for key, _ in list(node.outputs.items()):
                    outputs.add_trait(key, traits.Any(node=node))
                    setattr(outputs, key, None)
                setattr(outputdict, node.name, outputs)
        return outputdict

    def _set_input(self, objekt, name, newvalue):
        """Trait callback function to update a node input
        """
        objekt.traits()[name].node.set_input(name, newvalue)

    def _set_node_input(self, node, param, source, sourceinfo):
        """Set inputs of a node given the edge connection"""
        if isinstance(sourceinfo, (str, bytes)):
            val = source.get_output(sourceinfo)
        elif isinstance(sourceinfo, tuple):
            if callable(sourceinfo[1]):
                val = sourceinfo[1](source.get_output(sourceinfo[0]),
                                    *sourceinfo[2:])
        newval = val
        if isinstance(val, TraitDictObject):
            newval = dict(val)
        if isinstance(val, TraitListObject):
            newval = val[:]
        logger.debug('setting node input: %s->%s', param, to_str(newval))
        node.set_input(param, deepcopy(newval))

    def _get_all_nodes(self):
        allnodes = []
        for node in self._graph.nodes():
            if isinstance(node, Workflow):
                allnodes.extend(node._get_all_nodes())
            else:
                allnodes.append(node)
        return allnodes

    def _has_node(self, wanted_node):
        for node in self._graph.nodes():
            if wanted_node == node:
                return True
            if isinstance(node, Workflow):
                if node._has_node(wanted_node):
                    return True
        return False

    def _create_flat_graph(self):
        """Make a simple DAG where no node is a workflow."""
        logger.debug('Creating flat graph for workflow: %s', self.name)
        workflowcopy = deepcopy(self)
        workflowcopy._generate_flatgraph()
        return workflowcopy._graph

    def _reset_hierarchy(self):
        """Reset the hierarchy on a graph
        """
        for node in self._graph.nodes():
            if isinstance(node, Workflow):
                node._reset_hierarchy()
                for innernode in node._graph.nodes():
                    innernode._hierarchy = '.'.join((self.name,
                                                     innernode._hierarchy))
            else:
                node._hierarchy = self.name

    def _generate_flatgraph(self):
        """Generate a graph containing only Nodes or MapNodes
        """
        import networkx as nx
        logger.debug('expanding workflow: %s', self)
        nodes2remove = []
        if not nx.is_directed_acyclic_graph(self._graph):
            raise Exception(('Workflow: %s is not a directed acyclic graph '
                             '(DAG)') % self.name)
        nodes = list(nx.topological_sort(self._graph))
        for node in nodes:
            logger.debug('processing node: %s', node)
            if isinstance(node, Workflow):
                nodes2remove.append(node)
                # use in_edges instead of in_edges_iter to allow
                # disconnections to take place properly. otherwise, the
                # edge dict is modified.
                # dj: added list() for networkx ver.2
                for u, _, d in list(
                        self._graph.in_edges(nbunch=node, data=True)):
                    logger.debug('in: connections-> %s', to_str(d['connect']))
                    for cd in deepcopy(d['connect']):
                        logger.debug("in: %s", to_str(cd))
                        dstnode = node._get_parameter_node(cd[1], subtype='in')
                        srcnode = u
                        srcout = cd[0]
                        dstin = cd[1].split('.')[-1]
                        logger.debug('in edges: %s %s %s %s', srcnode, srcout,
                                     dstnode, dstin)
                        self.disconnect(u, cd[0], node, cd[1])
                        self.connect(srcnode, srcout, dstnode, dstin)
                # do not use out_edges_iter for reasons stated in in_edges
                # dj: for ver 2 use list(out_edges)
                for _, v, d in list(
                        self._graph.out_edges(nbunch=node, data=True)):
                    logger.debug('out: connections-> %s', to_str(d['connect']))
                    for cd in deepcopy(d['connect']):
                        logger.debug("out: %s", to_str(cd))
                        dstnode = v
                        if isinstance(cd[0], tuple):
                            parameter = cd[0][0]
                        else:
                            parameter = cd[0]
                        srcnode = node._get_parameter_node(
                            parameter, subtype='out')
                        if isinstance(cd[0], tuple):
                            srcout = list(cd[0])
                            srcout[0] = parameter.split('.')[-1]
                            srcout = tuple(srcout)
                        else:
                            srcout = parameter.split('.')[-1]
                        dstin = cd[1]
                        logger.debug('out edges: %s %s %s %s', srcnode, srcout,
                                     dstnode, dstin)
                        self.disconnect(node, cd[0], v, cd[1])
                        self.connect(srcnode, srcout, dstnode, dstin)
                # expand the workflow node
                # logger.debug('expanding workflow: %s', node)
                node._generate_flatgraph()
                for innernode in node._graph.nodes():
                    innernode._hierarchy = '.'.join((self.name,
                                                     innernode._hierarchy))
                self._graph.add_nodes_from(node._graph.nodes())
                self._graph.add_edges_from(node._graph.edges(data=True))
        if nodes2remove:
            self._graph.remove_nodes_from(nodes2remove)
        logger.debug('finished expanding workflow: %s', self)

    def _get_dot(self,
                 prefix=None,
                 hierarchy=None,
                 colored=False,
                 simple_form=True,
                 level=0):
        """Create a dot file with connection info
        """
        import networkx as nx
        if prefix is None:
            prefix = '  '
        if hierarchy is None:
            hierarchy = []
        colorset = [
            '#FFFFC8',  # Y
            '#0000FF',
            '#B4B4FF',
            '#E6E6FF',  # B
            '#FF0000',
            '#FFB4B4',
            '#FFE6E6',  # R
            '#00A300',
            '#B4FFB4',
            '#E6FFE6',  # G
            '#0000FF',
            '#B4B4FF'
        ]  # loop B
        if level > len(colorset) - 2:
            level = 3  # Loop back to blue

        dotlist = ['%slabel="%s";' % (prefix, self.name)]
        for node in nx.topological_sort(self._graph):
            fullname = '.'.join(hierarchy + [node.fullname])
            nodename = fullname.replace('.', '_')
            if not isinstance(node, Workflow):
                node_class_name = get_print_name(node, simple_form=simple_form)
                if not simple_form:
                    node_class_name = '.'.join(node_class_name.split('.')[1:])
                if hasattr(node, 'iterables') and node.iterables:
                    dotlist.append(('%s[label="%s", shape=box3d,'
                                    'style=filled, color=black, colorscheme'
                                    '=greys7 fillcolor=2];') %
                                   (nodename, node_class_name))
                else:
                    if colored:
                        dotlist.append(
                            ('%s[label="%s", style=filled,'
                             ' fillcolor="%s"];') % (nodename, node_class_name,
                                                     colorset[level]))
                    else:
                        dotlist.append(('%s[label="%s"];') % (nodename,
                                                              node_class_name))

        for node in nx.topological_sort(self._graph):
            if isinstance(node, Workflow):
                fullname = '.'.join(hierarchy + [node.fullname])
                nodename = fullname.replace('.', '_')
                dotlist.append('subgraph cluster_%s {' % nodename)
                if colored:
                    dotlist.append(prefix + prefix + 'edge [color="%s"];' %
                                   (colorset[level + 1]))
                    dotlist.append(prefix + prefix + 'style=filled;')
                    dotlist.append(prefix + prefix + 'fillcolor="%s";' %
                                   (colorset[level + 2]))
                dotlist.append(
                    node._get_dot(
                        prefix=prefix + prefix,
                        hierarchy=hierarchy + [self.name],
                        colored=colored,
                        simple_form=simple_form,
                        level=level + 3))
                dotlist.append('}')
            else:
                for subnode in self._graph.successors(node):
                    if node._hierarchy != subnode._hierarchy:
                        continue
                    if not isinstance(subnode, Workflow):
                        nodefullname = '.'.join(hierarchy + [node.fullname])
                        subnodefullname = '.'.join(
                            hierarchy + [subnode.fullname])
                        nodename = nodefullname.replace('.', '_')
                        subnodename = subnodefullname.replace('.', '_')
                        for _ in self._graph.get_edge_data(node,
                                                           subnode)['connect']:
                            dotlist.append('%s -> %s;' % (nodename,
                                                          subnodename))
                        logger.debug('connection: %s', dotlist[-1])
        # add between workflow connections
        for u, v, d in self._graph.edges(data=True):
            uname = '.'.join(hierarchy + [u.fullname])
            vname = '.'.join(hierarchy + [v.fullname])
            for src, dest in d['connect']:
                uname1 = uname
                vname1 = vname
                if isinstance(src, tuple):
                    srcname = src[0]
                else:
                    srcname = src
                if '.' in srcname:
                    uname1 += '.' + '.'.join(srcname.split('.')[:-1])
                if '.' in dest and '@' not in dest:
                    if not isinstance(v, Workflow):
                        if 'datasink' not in \
                           str(v._interface.__class__).lower():
                            vname1 += '.' + '.'.join(dest.split('.')[:-1])
                    else:
                        vname1 += '.' + '.'.join(dest.split('.')[:-1])
                if uname1.split('.')[:-1] != vname1.split('.')[:-1]:
                    dotlist.append('%s -> %s;' % (uname1.replace('.', '_'),
                                                  vname1.replace('.', '_')))
                    logger.debug('cross connection: %s', dotlist[-1])
        return ('\n' + prefix).join(dotlist)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import ConvertWarp


def test_ConvertWarp_inputs():
    input_map = dict(
        abswarp=dict(
            argstr='--abs',
            xor=['relwarp'],
        ),
        args=dict(argstr='%s', ),
        cons_jacobian=dict(argstr='--constrainj', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        jacobian_max=dict(argstr='--jmax=%f', ),
        jacobian_min=dict(argstr='--jmin=%f', ),
        midmat=dict(
            argstr='--midmat=%s',
            extensions=None,
        ),
        out_abswarp=dict(
            argstr='--absout',
            xor=['out_relwarp'],
        ),
        out_file=dict(
            argstr='--out=%s',
            extensions=None,
            name_source=['reference'],
            name_template='%s_concatwarp',
            output_name='out_file',
            position=-1,
        ),
        out_relwarp=dict(
            argstr='--relout',
            xor=['out_abswarp'],
        ),
        output_type=dict(),
        postmat=dict(
            argstr='--postmat=%s',
            extensions=None,
        ),
        premat=dict(
            argstr='--premat=%s',
            extensions=None,
        ),
        reference=dict(
            argstr='--ref=%s',
            extensions=None,
            mandatory=True,
            position=1,
        ),
        relwarp=dict(
            argstr='--rel',
            xor=['abswarp'],
        ),
        shift_direction=dict(
            argstr='--shiftdir=%s',
            requires=['shift_in_file'],
        ),
        shift_in_file=dict(
            argstr='--shiftmap=%s',
            extensions=None,
        ),
        warp1=dict(
            argstr='--warp1=%s',
            extensions=None,
        ),
        warp2=dict(
            argstr='--warp2=%s',
            extensions=None,
        ),
    )
    inputs = ConvertWarp.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ConvertWarp_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = ConvertWarp.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..minc import VolSymm


def test_VolSymm_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        clobber=dict(
            argstr='-clobber',
            usedefault=True,
        ),
        config_file=dict(
            argstr='-config_file %s',
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fit_linear=dict(argstr='-linear', ),
        fit_nonlinear=dict(argstr='-nonlinear', ),
        input_file=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=-3,
        ),
        input_grid_files=dict(),
        nofit=dict(argstr='-nofit', ),
        output_file=dict(
            argstr='%s',
            extensions=None,
            genfile=True,
            hash_files=False,
            name_source=['input_file'],
            name_template='%s_vol_symm.mnc',
            position=-1,
        ),
        trans_file=dict(
            argstr='%s',
            extensions=None,
            genfile=True,
            hash_files=False,
            keep_extension=False,
            name_source=['input_file'],
            name_template='%s_vol_symm.xfm',
            position=-2,
        ),
        verbose=dict(argstr='-verbose', ),
        x=dict(argstr='-x', ),
        y=dict(argstr='-y', ),
        z=dict(argstr='-z', ),
    )
    inputs = VolSymm.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_VolSymm_outputs():
    output_map = dict(
        output_file=dict(extensions=None, ),
        output_grid=dict(extensions=None, ),
        trans_file=dict(extensions=None, ),
    )
    outputs = VolSymm.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from .featuredetection import SphericalCoordinateGeneration
from .landmarkscompare import LandmarksCompare
from .generateaveragelmkfile import GenerateAverageLmkFile



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..registration import MRICoreg


def test_MRICoreg_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        brute_force_limit=dict(
            argstr='--bf-lim %g',
            xor=['no_brute_force'],
        ),
        brute_force_samples=dict(
            argstr='--bf-nsamp %d',
            xor=['no_brute_force'],
        ),
        conform_reference=dict(argstr='--conf-ref', ),
        dof=dict(argstr='--dof %d', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ftol=dict(argstr='--ftol %e', ),
        initial_rotation=dict(argstr='--rot %g %g %g', ),
        initial_scale=dict(argstr='--scale %g %g %g', ),
        initial_shear=dict(argstr='--shear %g %g %g', ),
        initial_translation=dict(argstr='--trans %g %g %g', ),
        linmintol=dict(argstr='--linmintol %e', ),
        max_iters=dict(argstr='--nitersmax %d', ),
        no_brute_force=dict(argstr='--no-bf', ),
        no_coord_dithering=dict(argstr='--no-coord-dither', ),
        no_cras0=dict(argstr='--no-cras0', ),
        no_intensity_dithering=dict(argstr='--no-intensity-dither', ),
        no_smooth=dict(argstr='--no-smooth', ),
        num_threads=dict(argstr='--threads %d', ),
        out_lta_file=dict(
            argstr='--lta %s',
            usedefault=True,
        ),
        out_params_file=dict(argstr='--params %s', ),
        out_reg_file=dict(argstr='--regdat %s', ),
        ref_fwhm=dict(argstr='--ref-fwhm', ),
        reference_file=dict(
            argstr='--ref %s',
            copyfile=False,
            extensions=None,
            mandatory=True,
            xor=['subject_id'],
        ),
        reference_mask=dict(
            argstr='--ref-mask %s',
            position=2,
        ),
        saturation_threshold=dict(argstr='--sat %g', ),
        sep=dict(argstr='--sep %s...', ),
        source_file=dict(
            argstr='--mov %s',
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        source_mask=dict(argstr='--mov-mask', ),
        source_oob=dict(argstr='--mov-oob', ),
        subject_id=dict(
            argstr='--s %s',
            mandatory=True,
            position=1,
            requires=['subjects_dir'],
            xor=['reference_file'],
        ),
        subjects_dir=dict(argstr='--sd %s', ),
    )
    inputs = MRICoreg.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_MRICoreg_outputs():
    output_map = dict(
        out_lta_file=dict(extensions=None, ),
        out_params_file=dict(extensions=None, ),
        out_reg_file=dict(extensions=None, ),
    )
    outputs = MRICoreg.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Nipype interface for seg_maths.

The maths module provides higher-level interfaces to some of the operations
that can be performed with the niftysegmaths (seg_maths) command-line program.

Examples
--------
See the docstrings of the individual classes for examples.
"""

import os

from ..base import (TraitedSpec, File, traits, isdefined, CommandLineInputSpec,
                    NipypeInterfaceError)
from .base import NiftySegCommand
from ..niftyreg.base import get_custom_path
from ...utils.filemanip import split_filename


class MathsInput(CommandLineInputSpec):
    """Input Spec for seg_maths interfaces."""
    in_file = File(
        position=2,
        argstr='%s',
        exists=True,
        mandatory=True,
        desc='image to operate on')

    out_file = File(
        name_source=['in_file'],
        name_template='%s',
        position=-2,
        argstr='%s',
        desc='image to write')

    desc = 'datatype to use for output (default uses input type)'
    output_datatype = traits.Enum(
        'float',
        'char',
        'int',
        'short',
        'double',
        'input',
        position=-3,
        argstr='-odt %s',
        desc=desc)


class MathsOutput(TraitedSpec):
    """Output Spec for seg_maths interfaces."""
    out_file = File(desc='image written after calculations')


class MathsCommand(NiftySegCommand):
    """
    Base Command Interface for seg_maths interfaces.

    The executable seg_maths enables the sequential execution of arithmetic
    operations, like multiplication (-mul), division (-div) or addition
    (-add), binarisation (-bin) or thresholding (-thr) operations and
    convolution by a Gaussian kernel (-smo). It also alows mathematical
    morphology based operations like dilation (-dil), erosion (-ero),
    connected components (-lconcomp) and hole filling (-fill), Euclidean
    (- euc) and geodesic (-geo) distance transforms, local image similarity
    metric calculation (-lncc and -lssd). Finally, it allows multiple
    operations over the dimensionality of the image, from merging 3D images
    together as a 4D image (-merge) or splitting (-split or -tp) 4D images
    into several 3D images, to estimating the maximum, minimum and average
    over all time-points, etc.
    """
    _cmd = get_custom_path('seg_maths', env_dir='NIFTYSEGDIR')
    input_spec = MathsInput
    output_spec = MathsOutput
    _suffix = '_maths'

    def _overload_extension(self, value, name=None):
        path, base, _ = split_filename(value)
        _, _, ext = split_filename(self.inputs.in_file)

        suffix = self._suffix
        if suffix != '_merged' and isdefined(self.inputs.operation):
            suffix = '_' + self.inputs.operation

        return os.path.join(path, '{0}{1}{2}'.format(base, suffix, ext))


class UnaryMathsInput(MathsInput):
    """Input Spec for seg_maths Unary operations."""
    operation = traits.Enum(
        'sqrt',
        'exp',
        'log',
        'recip',
        'abs',
        'bin',
        'otsu',
        'lconcomp',
        'concomp6',
        'concomp26',
        'fill',
        'euc',
        'tpmax',
        'tmean',
        'tmax',
        'tmin',
        'splitlab',
        'removenan',
        'isnan',
        'subsamp2',
        'scl',
        '4to5',
        'range',
        argstr='-%s',
        position=4,
        mandatory=True,
        desc='operation to perform')


class UnaryMaths(MathsCommand):
    """Interface for executable seg_maths from NiftySeg platform.

    Interface to use any unary mathematical operations that can be performed

    with the seg_maths command-line program.

    See below for those operations::

    sqrt - Square root of the image).

    exp - Exponential root of the image.

    log - Log of the image.

    recip - Reciprocal (1/I) of the image.

    abs - Absolute value of the image.

    bin - Binarise the image.

    otsu - Otsu thresholding of the current image.

    lconcomp - Take the largest connected component

    concomp6 - Label the different connected components with a 6NN kernel

    concomp26 - Label the different connected components with a 26NN kernel

    fill - Fill holes in binary object (e.g. fill ventricle in brain mask).

    euc - Euclidean distance trasnform

    tpmax - Get the time point with the highest value (binarise 4D \
probabilities)

    tmean - Mean value of all time points.

    tmax - Max value of all time points.

    tmin - Mean value of all time points.

    splitlab - Split the integer labels into multiple timepoints

    removenan - Remove all NaNs and replace then with 0

    isnan - Binary image equal to 1 if the value is NaN and 0 otherwise

    subsamp2 - Subsample the image by 2 using NN sampling (qform and sform \
scaled)

    scl  - Reset scale and slope info.

    4to5 - Flip the 4th and 5th dimension.

    range - Reset the image range to the min max.

    `Source code <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg>`_ |
    `Documentation <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg_documentation>`_

    Examples
    --------
    >>> import copy
    >>> from nipype.interfaces import niftyseg
    >>> unary = niftyseg.UnaryMaths()
    >>> unary.inputs.output_datatype = 'float'
    >>> unary.inputs.in_file = 'im1.nii'
    >>> # Test sqrt operation
    >>> unary_sqrt = copy.deepcopy(unary)
    >>> unary_sqrt.inputs.operation = 'sqrt'
    >>> unary_sqrt.cmdline
    'seg_maths im1.nii -sqrt -odt float im1_sqrt.nii'
    >>> unary_sqrt.run()  # doctest: +SKIP
    >>> # Test sqrt operation
    >>> unary_abs = copy.deepcopy(unary)
    >>> unary_abs.inputs.operation = 'abs'
    >>> unary_abs.cmdline
    'seg_maths im1.nii -abs -odt float im1_abs.nii'
    >>> unary_abs.run()  # doctest: +SKIP
    >>> # Test bin operation
    >>> unary_bin = copy.deepcopy(unary)
    >>> unary_bin.inputs.operation = 'bin'
    >>> unary_bin.cmdline
    'seg_maths im1.nii -bin -odt float im1_bin.nii'
    >>> unary_bin.run()  # doctest: +SKIP
    >>> # Test otsu operation
    >>> unary_otsu = copy.deepcopy(unary)
    >>> unary_otsu.inputs.operation = 'otsu'
    >>> unary_otsu.cmdline
    'seg_maths im1.nii -otsu -odt float im1_otsu.nii'
    >>> unary_otsu.run()  # doctest: +SKIP
    >>> # Test isnan operation
    >>> unary_isnan = copy.deepcopy(unary)
    >>> unary_isnan.inputs.operation = 'isnan'
    >>> unary_isnan.cmdline
    'seg_maths im1.nii -isnan -odt float im1_isnan.nii'
    >>> unary_isnan.run()  # doctest: +SKIP

    """
    input_spec = UnaryMathsInput


class BinaryMathsInput(MathsInput):
    """Input Spec for seg_maths Binary operations."""
    operation = traits.Enum(
        'mul',
        'div',
        'add',
        'sub',
        'pow',
        'thr',
        'uthr',
        'smo',
        'edge',
        'sobel3',
        'sobel5',
        'min',
        'smol',
        'geo',
        'llsnorm',
        'masknan',
        'hdr_copy',
        'splitinter',
        mandatory=True,
        argstr='-%s',
        position=4,
        desc='operation to perform')

    operand_file = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=5,
        xor=['operand_value', 'operand_str'],
        desc='second image to perform operation with')

    operand_value = traits.Float(
        argstr='%.8f',
        mandatory=True,
        position=5,
        xor=['operand_file', 'operand_str'],
        desc='float value to perform operation with')

    desc = 'string value to perform operation splitinter'
    operand_str = traits.Enum(
        'x',
        'y',
        'z',
        argstr='%s',
        mandatory=True,
        position=5,
        xor=['operand_value', 'operand_file'],
        desc=desc)


class BinaryMaths(MathsCommand):
    """Interface for executable seg_maths from NiftySeg platform.

    Interface to use any binary mathematical operations that can be performed

    with the seg_maths command-line program.

    See below for those operations::

    mul - <float/file> - Multiply image <float> value or by other image.

    div - <float/file> - Divide image by <float> or by other image.

    add - <float/file> - Add image by <float> or by other image.

    sub - <float/file> - Subtract image by <float> or by other image.

    pow - <float> - Image to the power of <float>.

    thr - <float> - Threshold the image below <float>.

    uthr - <float> - Threshold image above <float>.

    smo - <float> - Gaussian smoothing by std <float> (in voxels and up to \
4-D).

    edge - <float> - Calculate the edges of the image using a threshold <\
float>.

    sobel3 - <float> - Calculate the edges of all timepoints using a Sobel \
filter with a 3x3x3 kernel and applying <float> gaussian smoothing.

    sobel5 - <float> - Calculate the edges of all timepoints using a Sobel \
filter with a 5x5x5 kernel and applying <float> gaussian smoothing.

    min - <file> - Get the min per voxel between <current> and <file>.

    smol - <float> - Gaussian smoothing of a 3D label image.

    geo - <float/file> - Geodesic distance according to the speed function \
<float/file>

    llsnorm  <file_norm> - Linear LS normalisation between current and \
<file_norm>

    masknan <file_norm> - Assign everything outside the mask (mask==0) \
with NaNs

    hdr_copy <file> - Copy header from working image to <file> and save in \
<output>.

    splitinter <x/y/z> - Split interleaved slices in direction <x/y/z> into \
separate time points

    `Source code <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg>`_ |
    `Documentation <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg_documentation>`_

    Examples
    --------
    >>> import copy
    >>> from nipype.interfaces import niftyseg
    >>> binary = niftyseg.BinaryMaths()
    >>> binary.inputs.in_file = 'im1.nii'
    >>> binary.inputs.output_datatype = 'float'
    >>> # Test sub operation
    >>> binary_sub = copy.deepcopy(binary)
    >>> binary_sub.inputs.operation = 'sub'
    >>> binary_sub.inputs.operand_file = 'im2.nii'
    >>> binary_sub.cmdline
    'seg_maths im1.nii -sub im2.nii -odt float im1_sub.nii'
    >>> binary_sub.run()  # doctest: +SKIP
    >>> # Test mul operation
    >>> binary_mul = copy.deepcopy(binary)
    >>> binary_mul.inputs.operation = 'mul'
    >>> binary_mul.inputs.operand_value = 2.0
    >>> binary_mul.cmdline
    'seg_maths im1.nii -mul 2.00000000 -odt float im1_mul.nii'
    >>> binary_mul.run()  # doctest: +SKIP
    >>> # Test llsnorm operation
    >>> binary_llsnorm = copy.deepcopy(binary)
    >>> binary_llsnorm.inputs.operation = 'llsnorm'
    >>> binary_llsnorm.inputs.operand_file = 'im2.nii'
    >>> binary_llsnorm.cmdline
    'seg_maths im1.nii -llsnorm im2.nii -odt float im1_llsnorm.nii'
    >>> binary_llsnorm.run()  # doctest: +SKIP
    >>> # Test splitinter operation
    >>> binary_splitinter = copy.deepcopy(binary)
    >>> binary_splitinter.inputs.operation = 'splitinter'
    >>> binary_splitinter.inputs.operand_str = 'z'
    >>> binary_splitinter.cmdline
    'seg_maths im1.nii -splitinter z -odt float im1_splitinter.nii'
    >>> binary_splitinter.run()  # doctest: +SKIP

    """
    input_spec = BinaryMathsInput

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for seg_maths."""
        if opt == 'operand_str' and self.inputs.operation != 'splitinter':
            err = 'operand_str set but with an operation different than \
"splitinter"'

            raise NipypeInterfaceError(err)

        if opt == 'operation':
            # Only float
            if val in [
                    'pow', 'thr', 'uthr', 'smo', 'edge', 'sobel3', 'sobel5',
                    'smol'
            ]:
                if not isdefined(self.inputs.operand_value):
                    err = 'operand_value not set for {0}.'.format(val)
                    raise NipypeInterfaceError(err)
            # only files
            elif val in ['min', 'llsnorm', 'masknan', 'hdr_copy']:
                if not isdefined(self.inputs.operand_file):
                    err = 'operand_file not set for {0}.'.format(val)
                    raise NipypeInterfaceError(err)
            # splitinter:
            elif val == 'splitinter':
                if not isdefined(self.inputs.operand_str):
                    err = 'operand_str not set for splitinter.'
                    raise NipypeInterfaceError(err)

        if opt == 'operand_value' and float(val) == 0.0:
            return '0'

        return super(BinaryMaths, self)._format_arg(opt, spec, val)

    def _overload_extension(self, value, name=None):
        if self.inputs.operation == 'hdr_copy':
            path, base, _ = split_filename(value)
            _, base, ext = split_filename(self.inputs.operand_file)
            suffix = self.inputs.operation
            return os.path.join(path, '{0}{1}{2}'.format(base, suffix, ext))
        else:
            return super(BinaryMaths, self)._overload_extension(value, name)


class BinaryMathsInputInteger(MathsInput):
    """Input Spec for seg_maths Binary operations that require integer."""
    operation = traits.Enum(
        'dil',
        'ero',
        'tp',
        'equal',
        'pad',
        'crop',
        mandatory=True,
        argstr='-%s',
        position=4,
        desc='operation to perform')

    operand_value = traits.Int(
        argstr='%d',
        mandatory=True,
        position=5,
        desc='int value to perform operation with')


class BinaryMathsInteger(MathsCommand):
    """Interface for executable seg_maths from NiftySeg platform.

    Interface to use any integer mathematical operations that can be performed

    with the seg_maths command-line program.

    See below for those operations:: (requiring integer values)

    equal - <int> - Get voxels equal to <int>

    dil - <int>  - Dilate the image <int> times (in voxels).

    ero - <int> - Erode the image <int> times (in voxels).

    tp - <int> - Extract time point <int>

    crop - <int> - Crop <int> voxels around each 3D volume.

    pad - <int> -  Pad <int> voxels with NaN value around each 3D volume.

    `Source code <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg>`_ |
    `Documentation <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg_documentation>`_

    Examples
    --------
    >>> import copy
    >>> from nipype.interfaces.niftyseg import BinaryMathsInteger
    >>> binaryi = BinaryMathsInteger()
    >>> binaryi.inputs.in_file = 'im1.nii'
    >>> binaryi.inputs.output_datatype = 'float'
    >>> # Test dil operation
    >>> binaryi_dil = copy.deepcopy(binaryi)
    >>> binaryi_dil.inputs.operation = 'dil'
    >>> binaryi_dil.inputs.operand_value = 2
    >>> binaryi_dil.cmdline
    'seg_maths im1.nii -dil 2 -odt float im1_dil.nii'
    >>> binaryi_dil.run()  # doctest: +SKIP
    >>> # Test dil operation
    >>> binaryi_ero = copy.deepcopy(binaryi)
    >>> binaryi_ero.inputs.operation = 'ero'
    >>> binaryi_ero.inputs.operand_value = 1
    >>> binaryi_ero.cmdline
    'seg_maths im1.nii -ero 1 -odt float im1_ero.nii'
    >>> binaryi_ero.run()  # doctest: +SKIP
    >>> # Test pad operation
    >>> binaryi_pad = copy.deepcopy(binaryi)
    >>> binaryi_pad.inputs.operation = 'pad'
    >>> binaryi_pad.inputs.operand_value = 4
    >>> binaryi_pad.cmdline
    'seg_maths im1.nii -pad 4 -odt float im1_pad.nii'
    >>> binaryi_pad.run()  # doctest: +SKIP

    """
    input_spec = BinaryMathsInputInteger


class TupleMathsInput(MathsInput):
    """Input Spec for seg_maths Tuple operations."""
    operation = traits.Enum(
        'lncc',
        'lssd',
        'lltsnorm',
        mandatory=True,
        argstr='-%s',
        position=4,
        desc='operation to perform')

    operand_file1 = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=5,
        xor=['operand_value1'],
        desc='image to perform operation 1 with')

    desc = 'float value to perform operation 1 with'
    operand_value1 = traits.Float(
        argstr='%.8f',
        mandatory=True,
        position=5,
        xor=['operand_file1'],
        desc=desc)

    operand_file2 = File(
        exists=True,
        argstr='%s',
        mandatory=True,
        position=6,
        xor=['operand_value2'],
        desc='image to perform operation 2 with')

    desc = 'float value to perform operation 2 with'
    operand_value2 = traits.Float(
        argstr='%.8f',
        mandatory=True,
        position=6,
        xor=['operand_file2'],
        desc=desc)


class TupleMaths(MathsCommand):
    """Interface for executable seg_maths from NiftySeg platform.

    Interface to use any tuple mathematical operations that can be performed

    with the seg_maths command-line program.

    See below for those operations::

    lncc <file> <std> Local CC between current img and <file> on a kernel \
with <std>

    lssd <file> <std> Local SSD between current img and <file> on a kernel \
with <std>

    lltsnorm <file_norm> <float>  Linear LTS normalisation assuming <float> \
percent outliers

    `Source code <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg>`_ |
    `Documentation <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg_documentation>`_

    Examples
    --------
    >>> import copy
    >>> from nipype.interfaces import niftyseg
    >>> tuple = niftyseg.TupleMaths()
    >>> tuple.inputs.in_file = 'im1.nii'
    >>> tuple.inputs.output_datatype = 'float'

    >>> # Test lncc operation
    >>> tuple_lncc = copy.deepcopy(tuple)
    >>> tuple_lncc.inputs.operation = 'lncc'
    >>> tuple_lncc.inputs.operand_file1 = 'im2.nii'
    >>> tuple_lncc.inputs.operand_value2 = 2.0
    >>> tuple_lncc.cmdline
    'seg_maths im1.nii -lncc im2.nii 2.00000000 -odt float im1_lncc.nii'
    >>> tuple_lncc.run()  # doctest: +SKIP

    >>> # Test lssd operation
    >>> tuple_lssd = copy.deepcopy(tuple)
    >>> tuple_lssd.inputs.operation = 'lssd'
    >>> tuple_lssd.inputs.operand_file1 = 'im2.nii'
    >>> tuple_lssd.inputs.operand_value2 = 1.0
    >>> tuple_lssd.cmdline
    'seg_maths im1.nii -lssd im2.nii 1.00000000 -odt float im1_lssd.nii'
    >>> tuple_lssd.run()  # doctest: +SKIP

    >>> # Test lltsnorm operation
    >>> tuple_lltsnorm = copy.deepcopy(tuple)
    >>> tuple_lltsnorm.inputs.operation = 'lltsnorm'
    >>> tuple_lltsnorm.inputs.operand_file1 = 'im2.nii'
    >>> tuple_lltsnorm.inputs.operand_value2 = 0.01
    >>> tuple_lltsnorm.cmdline
    'seg_maths im1.nii -lltsnorm im2.nii 0.01000000 -odt float \
im1_lltsnorm.nii'
    >>> tuple_lltsnorm.run()  # doctest: +SKIP
    """
    input_spec = TupleMathsInput


class MergeInput(MathsInput):
    """Input Spec for seg_maths merge operation."""
    dimension = traits.Int(
        mandatory=True, desc='Dimension to merge the images.')

    desc = 'List of images to merge to the working image <input>.'
    merge_files = traits.List(
        File(exists=True), argstr='%s', mandatory=True, position=4, desc=desc)


class Merge(MathsCommand):
    """Interface for executable seg_maths from NiftySeg platform.

    Interface to use the merge operation that can be performed

    with the seg_maths command-line program.

    See below for this option::

    merge  <i> <d> <files>  Merge <i> images and the working image in the \
<d> dimension

    `Source code <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg>`_ |
    `Documentation <http://cmictig.cs.ucl.ac.uk/wiki/index.php/NiftySeg_documentation>`_

    Examples
    --------
    >>> from nipype.interfaces import niftyseg
    >>> node = niftyseg.Merge()
    >>> node.inputs.in_file = 'im1.nii'
    >>> files = ['im2.nii', 'im3.nii']
    >>> node.inputs.merge_files = files
    >>> node.inputs.dimension = 2
    >>> node.inputs.output_datatype = 'float'
    >>> node.cmdline
    'seg_maths im1.nii -merge 2 2 im2.nii im3.nii -odt float im1_merged.nii'

    """
    input_spec = MergeInput
    _suffix = '_merged'

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for seg_maths."""
        if opt == 'merge_files':
            return "-merge %d %d %s" % (len(val), self.inputs.dimension,
                                        ' '.join(val))

        return super(Merge, self)._format_arg(opt, spec, val)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..specialized import BinaryMaskEditorBasedOnLandmarks


def test_BinaryMaskEditorBasedOnLandmarks_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputBinaryVolume=dict(
            argstr='--inputBinaryVolume %s',
            extensions=None,
        ),
        inputLandmarkNames=dict(
            argstr='--inputLandmarkNames %s',
            sep=',',
        ),
        inputLandmarkNamesForObliquePlane=dict(
            argstr='--inputLandmarkNamesForObliquePlane %s',
            sep=',',
        ),
        inputLandmarksFilename=dict(
            argstr='--inputLandmarksFilename %s',
            extensions=None,
        ),
        outputBinaryVolume=dict(
            argstr='--outputBinaryVolume %s',
            hash_files=False,
        ),
        setCutDirectionForLandmark=dict(
            argstr='--setCutDirectionForLandmark %s',
            sep=',',
        ),
        setCutDirectionForObliquePlane=dict(
            argstr='--setCutDirectionForObliquePlane %s',
            sep=',',
        ),
    )
    inputs = BinaryMaskEditorBasedOnLandmarks.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_BinaryMaskEditorBasedOnLandmarks_outputs():
    output_map = dict(outputBinaryVolume=dict(extensions=None, ), )
    outputs = BinaryMaskEditorBasedOnLandmarks.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import SurfaceTransform


def test_SurfaceTransform_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        hemi=dict(
            argstr='--hemi %s',
            mandatory=True,
        ),
        out_file=dict(
            argstr='--tval %s',
            extensions=None,
            genfile=True,
        ),
        reshape=dict(argstr='--reshape', ),
        reshape_factor=dict(argstr='--reshape-factor', ),
        source_annot_file=dict(
            argstr='--sval-annot %s',
            extensions=None,
            mandatory=True,
            xor=['source_file'],
        ),
        source_file=dict(
            argstr='--sval %s',
            extensions=None,
            mandatory=True,
            xor=['source_annot_file'],
        ),
        source_subject=dict(
            argstr='--srcsubject %s',
            mandatory=True,
        ),
        source_type=dict(
            argstr='--sfmt %s',
            requires=['source_file'],
        ),
        subjects_dir=dict(),
        target_ico_order=dict(argstr='--trgicoorder %d', ),
        target_subject=dict(
            argstr='--trgsubject %s',
            mandatory=True,
        ),
        target_type=dict(argstr='--tfmt %s', ),
    )
    inputs = SurfaceTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_SurfaceTransform_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = SurfaceTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..io import JSONFileSink


def test_JSONFileSink_inputs():
    input_map = dict(
        _outputs=dict(usedefault=True, ),
        in_dict=dict(usedefault=True, ),
        out_file=dict(extensions=None, ),
    )
    inputs = JSONFileSink.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_JSONFileSink_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = JSONFileSink.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Local serial workflow execution
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

import os
from .base import (PluginBase, logger, report_crash, report_nodes_not_run,
                   str2bool)
from ..engine.utils import topological_sort


class LinearPlugin(PluginBase):
    """Execute workflow in series
    """

    def run(self, graph, config, updatehash=False):
        """Executes a pre-defined pipeline in a serial order.

        Parameters
        ----------

        graph : networkx digraph
            defines order of execution
        """
        import networkx as nx
        try:
            dfs_preorder = nx.dfs_preorder
        except AttributeError:
            dfs_preorder = nx.dfs_preorder_nodes

        if not isinstance(graph, nx.DiGraph):
            raise ValueError('Input must be a networkx digraph object')
        logger.info("Running serially.")
        old_wd = os.getcwd()
        notrun = []
        donotrun = []
        nodes, _ = topological_sort(graph)
        for node in nodes:
            endstatus = 'end'
            try:
                if node in donotrun:
                    continue
                if self._status_callback:
                    self._status_callback(node, 'start')
                node.run(updatehash=updatehash)
            except:
                endstatus = 'exception'
                # bare except, but i really don't know where a
                # node might fail
                crashfile = report_crash(node)
                if str2bool(config['execution']['stop_on_first_crash']):
                    raise
                # remove dependencies from queue
                subnodes = [s for s in dfs_preorder(graph, node)]
                notrun.append({'node': node, 'dependents': subnodes,
                               'crashfile': crashfile})
                donotrun.extend(subnodes)
                # Delay raising the crash until we cleaned the house
                if str2bool(config['execution']['stop_on_first_crash']):
                    os.chdir(old_wd)  # Return wherever we were before
                    report_nodes_not_run(notrun)  # report before raising
                    raise
            finally:
                if self._status_callback:
                    self._status_callback(node, endstatus)

        os.chdir(old_wd)  # Return wherever we were before
        report_nodes_not_run(notrun)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..resampling import WarpImageMultiTransform


def test_WarpImageMultiTransform_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        dimension=dict(
            argstr='%d',
            position=1,
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        input_image=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=2,
        ),
        invert_affine=dict(),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_postfix=dict(
            extensions=None,
            hash_files=False,
            usedefault=True,
            xor=['output_image'],
        ),
        output_image=dict(
            argstr='%s',
            extensions=None,
            genfile=True,
            hash_files=False,
            position=3,
            xor=['out_postfix'],
        ),
        reference_image=dict(
            argstr='-R %s',
            extensions=None,
            xor=['tightest_box'],
        ),
        reslice_by_header=dict(argstr='--reslice-by-header', ),
        tightest_box=dict(
            argstr='--tightest-bounding-box',
            xor=['reference_image'],
        ),
        transformation_series=dict(
            argstr='%s',
            mandatory=True,
            position=-1,
        ),
        use_bspline=dict(argstr='--use-BSpline', ),
        use_nearest=dict(argstr='--use-NN', ),
    )
    inputs = WarpImageMultiTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_WarpImageMultiTransform_outputs():
    output_map = dict(output_image=dict(extensions=None, ), )
    outputs = WarpImageMultiTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
"""Parallel workflow execution via OAR http://oar.imag.fr
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from builtins import str, open
import os
import stat
from time import sleep
import subprocess
import simplejson as json

from ... import logging
from ...interfaces.base import CommandLine
from .base import SGELikeBatchManagerBase, logger
iflogger = logging.getLogger('nipype.interface')


class OARPlugin(SGELikeBatchManagerBase):
    """Execute using OAR

    The plugin_args input to run can be used to control the OAR execution.
    Currently supported options are:

    - template : template to use for batch job submission
    - oarsub_args : arguments to be prepended to the job execution
                    script in the oarsub call
    - max_jobname_len: maximum length of the job name.  Default 15.

    """

    # Addtional class variables
    _max_jobname_len = 15
    _oarsub_args = ''

    def __init__(self, **kwargs):
        template = """
# oarsub -J
        """
        self._retry_timeout = 2
        self._max_tries = 2
        self._max_jobname_length = 15
        if 'plugin_args' in kwargs and kwargs['plugin_args']:
            if 'oarsub_args' in kwargs['plugin_args']:
                self._oarsub_args = kwargs['plugin_args']['oarsub_args']
            if 'retry_timeout' in kwargs['plugin_args']:
                self._retry_timeout = kwargs['plugin_args']['retry_timeout']
            if 'max_tries' in kwargs['plugin_args']:
                self._max_tries = kwargs['plugin_args']['max_tries']
            if 'max_jobname_len' in kwargs['plugin_args']:
                self._max_jobname_len = \
                    kwargs['plugin_args']['max_jobname_len']
        super(OARPlugin, self).__init__(template, **kwargs)

    def _is_pending(self, taskid):
        #  subprocess.Popen requires taskid to be a string
        proc = subprocess.Popen(
            ['oarstat', '-J', '-s', '-j', taskid],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        o, e = proc.communicate()
        parsed_result = json.loads(o)[taskid].lower()
        is_pending = (('error' not in parsed_result)
                      and ('terminated' not in parsed_result))
        return is_pending

    def _submit_batchtask(self, scriptfile, node):
        cmd = CommandLine(
            'oarsub',
            environ=dict(os.environ),
            resource_monitor=False,
            terminal_output='allatonce')
        path = os.path.dirname(scriptfile)
        oarsubargs = ''
        if self._oarsub_args:
            oarsubargs = self._oarsub_args
        if 'oarsub_args' in node.plugin_args:
            if ('overwrite' in node.plugin_args
                    and node.plugin_args['overwrite']):
                oarsubargs = node.plugin_args['oarsub_args']
            else:
                oarsubargs += (" " + node.plugin_args['oarsub_args'])

        if node._hierarchy:
            jobname = '.'.join((dict(os.environ)['LOGNAME'], node._hierarchy,
                                node._id))
        else:
            jobname = '.'.join((dict(os.environ)['LOGNAME'], node._id))
        jobnameitems = jobname.split('.')
        jobnameitems.reverse()
        jobname = '.'.join(jobnameitems)
        jobname = jobname[0:self._max_jobname_len]

        if '-O' not in oarsubargs:
            oarsubargs = '%s -O %s' % (oarsubargs,
                                       os.path.join(path, jobname + '.stdout'))
        if '-E' not in oarsubargs:
            oarsubargs = '%s -E %s' % (oarsubargs,
                                       os.path.join(path, jobname + '.stderr'))
        if '-J' not in oarsubargs:
            oarsubargs = '%s -J' % (oarsubargs)

        os.chmod(scriptfile, stat.S_IEXEC | stat.S_IREAD | stat.S_IWRITE)
        cmd.inputs.args = '%s -n %s -S %s' % (oarsubargs, jobname, scriptfile)

        oldlevel = iflogger.level
        iflogger.setLevel(logging.getLevelName('CRITICAL'))
        tries = 0
        while True:
            try:
                result = cmd.run()
            except Exception as e:
                if tries < self._max_tries:
                    tries += 1
                    sleep(self._retry_timeout)
                    # sleep 2 seconds and try again.
                else:
                    iflogger.setLevel(oldlevel)
                    raise RuntimeError('\n'.join((('Could not submit OAR task'
                                                   ' for node %s') % node._id,
                                                  str(e))))
            else:
                break
        iflogger.setLevel(oldlevel)
        # retrieve OAR taskid

        o = ''
        add = False
        for line in result.runtime.stdout.splitlines():
            if line.strip().startswith('{'):
                add = True
            if add:
                o += line + '\n'
            if line.strip().startswith('}'):
                break
        taskid = json.loads(o)['job_id']
        self._pending[taskid] = node.output_dir()
        logger.debug('submitted OAR task: %s for node %s' % (taskid, node._id))
        return taskid



import os
import pytest
import shutil

from nipype.interfaces.dcm2nii import Dcm2niix
no_dcm2niix = not bool(Dcm2niix().version)
no_datalad = False
try:
    from datalad import api # to pull and grab data
    from datalad.support.exceptions import IncompleteResultsError
except ImportError:
    no_datalad = True

DICOM_DIR = 'http://datasets-tests.datalad.org/dicoms/dcm2niix-tests'


@pytest.fixture
def fetch_data():
    def _fetch_data(datadir, dicoms):
        try:
            """Fetches some test DICOMs using datalad"""
            api.install(path=datadir, source=DICOM_DIR)
            data = os.path.join(datadir, dicoms)
            api.get(path=data)
        except IncompleteResultsError as exc:
            pytest.skip("Failed to fetch test data: %s" % str(exc))
        return data
    return _fetch_data

@pytest.mark.skipif(no_datalad, reason="Datalad required")
@pytest.mark.skipif(no_dcm2niix, reason="Dcm2niix required")
@pytest.mark.xfail(reason="Intermittent failures. Let's come back to this later.")
def test_dcm2niix_dti(fetch_data, tmpdir):
    tmpdir.chdir()
    datadir = tmpdir.mkdir('data').strpath
    dicoms = fetch_data(datadir, 'Siemens_Sag_DTI_20160825_145811')

    def assert_dti(res):
        "Some assertions we will make"
        assert res.outputs.converted_files
        assert res.outputs.bvals
        assert res.outputs.bvecs
        outputs = [y for x,y in res.outputs.get().items()]
        if res.inputs.get('bids_format'):
            # ensure all outputs are of equal lengths
            assert len(set(map(len, outputs))) == 1
        else:
            assert not res.outputs.bids

    dcm = Dcm2niix()
    dcm.inputs.source_dir = dicoms
    dcm.inputs.out_filename = '%u%z'
    assert_dti(dcm.run())

    # now run specifying output directory and removing BIDS option
    outdir = tmpdir.mkdir('conversion').strpath
    dcm.inputs.output_dir = outdir
    dcm.inputs.bids_format = False
    assert_dti(dcm.run())



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import ApplyDeformations


def test_ApplyDeformations_inputs():
    input_map = dict(
        deformation_field=dict(
            extensions=None,
            field='comp{1}.def',
            mandatory=True,
        ),
        in_files=dict(
            field='fnames',
            mandatory=True,
        ),
        interp=dict(field='interp', ),
        matlab_cmd=dict(),
        mfile=dict(usedefault=True, ),
        paths=dict(),
        reference_volume=dict(
            extensions=['.hdr', '.img', '.img.gz', '.nii'],
            field='comp{2}.id.space',
            mandatory=True,
        ),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver='8',
            usedefault=True,
        ),
    )
    inputs = ApplyDeformations.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ApplyDeformations_outputs():
    output_map = dict(out_files=dict(), )
    outputs = ApplyDeformations.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..utils import Eval


def test_Eval_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        expr=dict(
            argstr='-expr "%s"',
            mandatory=True,
            position=3,
        ),
        in_file_a=dict(
            argstr='-a %s',
            extensions=None,
            mandatory=True,
            position=0,
        ),
        in_file_b=dict(
            argstr='-b %s',
            extensions=None,
            position=1,
        ),
        in_file_c=dict(
            argstr='-c %s',
            extensions=None,
            position=2,
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        other=dict(
            argstr='',
            extensions=None,
        ),
        out1D=dict(argstr='-1D', ),
        out_file=dict(
            argstr='-prefix %s',
            extensions=None,
            name_source='in_file_a',
            name_template='%s_calc',
        ),
        outputtype=dict(),
        single_idx=dict(),
        start_idx=dict(requires=['stop_idx'], ),
        stop_idx=dict(requires=['start_idx'], ),
    )
    inputs = Eval.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Eval_outputs():
    output_map = dict(out_file=dict(extensions=None, ), )
    outputs = Eval.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..specialized import BRAINSDemonWarp


def test_BRAINSDemonWarp_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        arrayOfPyramidLevelIterations=dict(
            argstr='--arrayOfPyramidLevelIterations %s',
            sep=',',
        ),
        backgroundFillValue=dict(argstr='--backgroundFillValue %d', ),
        checkerboardPatternSubdivisions=dict(
            argstr='--checkerboardPatternSubdivisions %s',
            sep=',',
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fixedBinaryVolume=dict(
            argstr='--fixedBinaryVolume %s',
            extensions=None,
        ),
        fixedVolume=dict(
            argstr='--fixedVolume %s',
            extensions=None,
        ),
        gradient_type=dict(argstr='--gradient_type %s', ),
        gui=dict(argstr='--gui ', ),
        histogramMatch=dict(argstr='--histogramMatch ', ),
        initializeWithDisplacementField=dict(
            argstr='--initializeWithDisplacementField %s',
            extensions=None,
        ),
        initializeWithTransform=dict(
            argstr='--initializeWithTransform %s',
            extensions=None,
        ),
        inputPixelType=dict(argstr='--inputPixelType %s', ),
        interpolationMode=dict(argstr='--interpolationMode %s', ),
        lowerThresholdForBOBF=dict(argstr='--lowerThresholdForBOBF %d', ),
        maskProcessingMode=dict(argstr='--maskProcessingMode %s', ),
        max_step_length=dict(argstr='--max_step_length %f', ),
        medianFilterSize=dict(
            argstr='--medianFilterSize %s',
            sep=',',
        ),
        minimumFixedPyramid=dict(
            argstr='--minimumFixedPyramid %s',
            sep=',',
        ),
        minimumMovingPyramid=dict(
            argstr='--minimumMovingPyramid %s',
            sep=',',
        ),
        movingBinaryVolume=dict(
            argstr='--movingBinaryVolume %s',
            extensions=None,
        ),
        movingVolume=dict(
            argstr='--movingVolume %s',
            extensions=None,
        ),
        neighborhoodForBOBF=dict(
            argstr='--neighborhoodForBOBF %s',
            sep=',',
        ),
        numberOfBCHApproximationTerms=dict(
            argstr='--numberOfBCHApproximationTerms %d', ),
        numberOfHistogramBins=dict(argstr='--numberOfHistogramBins %d', ),
        numberOfMatchPoints=dict(argstr='--numberOfMatchPoints %d', ),
        numberOfPyramidLevels=dict(argstr='--numberOfPyramidLevels %d', ),
        numberOfThreads=dict(argstr='--numberOfThreads %d', ),
        outputCheckerboardVolume=dict(
            argstr='--outputCheckerboardVolume %s',
            hash_files=False,
        ),
        outputDebug=dict(argstr='--outputDebug ', ),
        outputDisplacementFieldPrefix=dict(
            argstr='--outputDisplacementFieldPrefix %s', ),
        outputDisplacementFieldVolume=dict(
            argstr='--outputDisplacementFieldVolume %s',
            hash_files=False,
        ),
        outputNormalized=dict(argstr='--outputNormalized ', ),
        outputPixelType=dict(argstr='--outputPixelType %s', ),
        outputVolume=dict(
            argstr='--outputVolume %s',
            hash_files=False,
        ),
        promptUser=dict(argstr='--promptUser ', ),
        registrationFilterType=dict(argstr='--registrationFilterType %s', ),
        seedForBOBF=dict(
            argstr='--seedForBOBF %s',
            sep=',',
        ),
        smoothDisplacementFieldSigma=dict(
            argstr='--smoothDisplacementFieldSigma %f', ),
        upFieldSmoothing=dict(argstr='--upFieldSmoothing %f', ),
        upperThresholdForBOBF=dict(argstr='--upperThresholdForBOBF %d', ),
        use_vanilla_dem=dict(argstr='--use_vanilla_dem ', ),
    )
    inputs = BRAINSDemonWarp.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_BRAINSDemonWarp_outputs():
    output_map = dict(
        outputCheckerboardVolume=dict(extensions=None, ),
        outputDisplacementFieldVolume=dict(extensions=None, ),
        outputVolume=dict(extensions=None, ),
    )
    outputs = BRAINSDemonWarp.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import pytest
import os

import nibabel as nb

from nipype.algorithms import misc
from nipype.utils.filemanip import fname_presuffix
from nipype.testing.fixtures import create_analyze_pair_file_in_directory
from nipype.utils import NUMPY_MMAP
from nipype.testing import example_data


def test_CreateNifti(create_analyze_pair_file_in_directory):

    filelist, outdir = create_analyze_pair_file_in_directory

    create_nifti = misc.CreateNifti()

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        create_nifti.run()

    # .inputs based parameters setting
    create_nifti.inputs.header_file = filelist[0]
    create_nifti.inputs.data_file = fname_presuffix(
        filelist[0], '', '.img', use_ext=False)

    result = create_nifti.run()

    assert os.path.exists(result.outputs.nifti_file)
    assert nb.load(result.outputs.nifti_file, mmap=NUMPY_MMAP)


def test_CalculateMedian(create_analyze_pair_file_in_directory):

    mean = misc.CalculateMedian()

    with pytest.raises(TypeError):
        mean.run()

    mean.inputs.in_files = example_data('ds003_sub-01_mc.nii.gz')
    eg = mean.run()

    assert os.path.exists(eg.outputs.median_files)
    assert nb.load(eg.outputs.median_files, mmap=NUMPY_MMAP)



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Common graph operations for execution
"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import open

import os
import getpass
from socket import gethostname
import sys
import uuid
from time import strftime
from traceback import format_exception

from ... import logging
from ...utils.filemanip import savepkl, crash2txt, makedirs, FileNotFoundError

logger = logging.getLogger('nipype.workflow')


def report_crash(node, traceback=None, hostname=None):
    """Writes crash related information to a file
    """
    name = node._id
    host = None
    traceback = traceback or format_exception(*sys.exc_info())

    try:
        result = node.result
    except FileNotFoundError:
        traceback += """

When creating this crashfile, the results file corresponding
to the node could not be found.""".splitlines(keepends=True)
    except Exception as exc:
        traceback += """

During the creation of this crashfile triggered by the above exception,
another exception occurred:\n\n{}.""".format(exc).splitlines(keepends=True)
    else:
        if getattr(result, 'runtime', None):
            if isinstance(result.runtime, list):
                host = result.runtime[0].hostname
            else:
                host = result.runtime.hostname

    # Try everything to fill in the host
    host = host or hostname or gethostname()
    logger.error('Node %s failed to run on host %s.', name, host)
    timeofcrash = strftime('%Y%m%d-%H%M%S')
    try:
        login_name = getpass.getuser()
    except KeyError:
        login_name = 'UID{:d}'.format(os.getuid())
    crashfile = 'crash-%s-%s-%s-%s' % (timeofcrash, login_name, name,
                                       str(uuid.uuid4()))
    crashdir = node.config['execution'].get('crashdump_dir', os.getcwd())

    makedirs(crashdir, exist_ok=True)
    crashfile = os.path.join(crashdir, crashfile)

    if node.config['execution']['crashfile_format'].lower() in ('text', 'txt', '.txt'):
        crashfile += '.txt'
    else:
        crashfile += '.pklz'

    logger.error('Saving crash info to %s\n%s', crashfile, ''.join(traceback))
    if crashfile.endswith('.txt'):
        crash2txt(crashfile, dict(node=node, traceback=traceback))
    else:
        savepkl(crashfile, dict(node=node, traceback=traceback),
                versioning=True)
    return crashfile


def report_nodes_not_run(notrun):
    """List nodes that crashed with crashfile info

    Optionally displays dependent nodes that weren't executed as a result of
    the crash.
    """
    if notrun:
        logger.info("***********************************")
        for info in notrun:
            logger.error("could not run node: %s" % '.'.join(
                (info['node']._hierarchy, info['node']._id)))
            logger.info("crashfile: %s" % info['crashfile'])
            logger.debug("The following dependent nodes were not run")
            for subnode in info['dependents']:
                logger.debug(subnode._id)
        logger.info("***********************************")
        raise RuntimeError(('Workflow did not execute cleanly. '
                            'Check log for details'))


def create_pyscript(node, updatehash=False, store_exception=True):
    # pickle node
    timestamp = strftime('%Y%m%d_%H%M%S')
    if node._hierarchy:
        suffix = '%s_%s_%s' % (timestamp, node._hierarchy, node._id)
        batch_dir = os.path.join(node.base_dir,
                                 node._hierarchy.split('.')[0], 'batch')
    else:
        suffix = '%s_%s' % (timestamp, node._id)
        batch_dir = os.path.join(node.base_dir, 'batch')
    if not os.path.exists(batch_dir):
        os.makedirs(batch_dir)
    pkl_file = os.path.join(batch_dir, 'node_%s.pklz' % suffix)
    savepkl(pkl_file, dict(node=node, updatehash=updatehash))
    mpl_backend = node.config["execution"]["matplotlib_backend"]
    # create python script to load and trap exception
    cmdstr = """import os
import sys

can_import_matplotlib = True #Silently allow matplotlib to be ignored
try:
    import matplotlib
    matplotlib.use('%s')
except ImportError:
    can_import_matplotlib = False
    pass

from nipype import config, logging
from nipype.utils.filemanip import loadpkl, savepkl
from socket import gethostname
from traceback import format_exception
info = None
pklfile = '%s'
batchdir = '%s'
from nipype.utils.filemanip import loadpkl, savepkl
try:
    if not sys.version_info < (2, 7):
        from collections import OrderedDict
    config_dict=%s
    config.update_config(config_dict)
    ## Only configure matplotlib if it was successfully imported,
    ## matplotlib is an optional component to nipype
    if can_import_matplotlib:
        config.update_matplotlib()
    logging.update_logging(config)
    traceback=None
    cwd = os.getcwd()
    info = loadpkl(pklfile)
    result = info['node'].run(updatehash=info['updatehash'])
except Exception as e:
    etype, eval, etr = sys.exc_info()
    traceback = format_exception(etype,eval,etr)
    if info is None or not os.path.exists(info['node'].output_dir()):
        result = None
        resultsfile = os.path.join(batchdir, 'crashdump_%s.pklz')
    else:
        result = info['node'].result
        resultsfile = os.path.join(info['node'].output_dir(),
                               'result_%%s.pklz'%%info['node'].name)
"""
    if store_exception:
        cmdstr += """
    savepkl(resultsfile, dict(result=result, hostname=gethostname(),
                              traceback=traceback))
"""
    else:
        cmdstr += """
    if info is None:
        savepkl(resultsfile, dict(result=result, hostname=gethostname(),
                              traceback=traceback))
    else:
        from nipype.pipeline.plugins.base import report_crash
        report_crash(info['node'], traceback, gethostname())
    raise Exception(e)
"""
    cmdstr = cmdstr % (mpl_backend, pkl_file, batch_dir, node.config, suffix)
    pyscript = os.path.join(batch_dir, 'pyscript_%s.py' % suffix)
    with open(pyscript, 'wt') as fp:
        fp.writelines(cmdstr)
    return pyscript



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import Warp


def test_Warp_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        deoblique=dict(argstr='-deoblique', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        gridset=dict(
            argstr='-gridset %s',
            extensions=None,
        ),
        in_file=dict(
            argstr='%s',
            copyfile=False,
            extensions=None,
            mandatory=True,
            position=-1,
        ),
        interp=dict(argstr='-%s', ),
        matparent=dict(
            argstr='-matparent %s',
            extensions=None,
        ),
        mni2tta=dict(argstr='-mni2tta', ),
        newgrid=dict(argstr='-newgrid %f', ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        oblique_parent=dict(
            argstr='-oblique_parent %s',
            extensions=None,
        ),
        out_file=dict(
            argstr='-prefix %s',
            extensions=None,
            keep_extension=True,
            name_source='in_file',
            name_template='%s_warp',
        ),
        outputtype=dict(),
        save_warp=dict(requires=['verbose'], ),
        tta2mni=dict(argstr='-tta2mni', ),
        verbose=dict(argstr='-verb', ),
        zpad=dict(argstr='-zpad %d', ),
    )
    inputs = Warp.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Warp_outputs():
    output_map = dict(
        out_file=dict(extensions=None, ),
        warp_file=dict(extensions=None, ),
    )
    outputs = Warp.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

from nipype.utils.docparse import reverse_opt_map, build_doc, insert_doc

foo_opt_map = {'outline': '-o', 'fun': '-f %.2f', 'flags': '%s'}

foo_doc = """Usage: foo infile outfile [opts]

Bunch of options:

  -o        something about an outline
  -f <f>    intensity of fun factor

Other stuff:
  -v        verbose

"""

fmtd_doc = """Parameters
----------
outline :
     something about an outline
fun :
     <f> intensity of fun factor

Others Parameters
-----------------
  -v        verbose"""


def test_rev_opt_map():
    map = {'-f': 'fun', '-o': 'outline'}
    rev_map = reverse_opt_map(foo_opt_map)
    assert rev_map == map


def test_build_doc():
    opts = reverse_opt_map(foo_opt_map)
    doc = build_doc(foo_doc, opts)
    assert doc == fmtd_doc


inserted_doc = """Parameters
----------
infile : str
    The name of the input file
outfile : str
    The name of the output file
outline :
     something about an outline
fun :
     <f> intensity of fun factor

Others Parameters
-----------------
  -v        verbose"""


def test_insert_doc():
    new_items = ['infile : str', '    The name of the input file']
    new_items.extend(['outfile : str', '    The name of the output file'])
    newdoc = insert_doc(fmtd_doc, new_items)
    assert newdoc == inserted_doc



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..model import MS_LDA


def test_MS_LDA_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        conform=dict(argstr='-conform', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        images=dict(
            argstr='%s',
            copyfile=False,
            mandatory=True,
            position=-1,
        ),
        label_file=dict(
            argstr='-label %s',
            extensions=None,
        ),
        lda_labels=dict(
            argstr='-lda %s',
            mandatory=True,
            sep=' ',
        ),
        mask_file=dict(
            argstr='-mask %s',
            extensions=None,
        ),
        shift=dict(argstr='-shift %d', ),
        subjects_dir=dict(),
        use_weights=dict(argstr='-W', ),
        vol_synth_file=dict(
            argstr='-synth %s',
            extensions=None,
            mandatory=True,
        ),
        weight_file=dict(
            argstr='-weight %s',
            extensions=None,
            mandatory=True,
        ),
    )
    inputs = MS_LDA.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_MS_LDA_outputs():
    output_map = dict(
        vol_synth_file=dict(extensions=None, ),
        weight_file=dict(extensions=None, ),
    )
    outputs = MS_LDA.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..tracking import DiffusionTensorStreamlineTrack


def test_DiffusionTensorStreamlineTrack_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        cutoff_value=dict(
            argstr='-cutoff %s',
            units='NA',
        ),
        desired_number_of_tracks=dict(argstr='-number %d', ),
        do_not_precompute=dict(argstr='-noprecomputed', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        exclude_file=dict(
            argstr='-exclude %s',
            extensions=None,
            xor=['exclude_file', 'exclude_spec'],
        ),
        exclude_spec=dict(
            argstr='-exclude %s',
            position=2,
            sep=',',
            units='mm',
            xor=['exclude_file', 'exclude_spec'],
        ),
        gradient_encoding_file=dict(
            argstr='-grad %s',
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        in_file=dict(
            argstr='%s',
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        include_file=dict(
            argstr='-include %s',
            extensions=None,
            xor=['include_file', 'include_spec'],
        ),
        include_spec=dict(
            argstr='-include %s',
            position=2,
            sep=',',
            units='mm',
            xor=['include_file', 'include_spec'],
        ),
        initial_cutoff_value=dict(
            argstr='-initcutoff %s',
            units='NA',
        ),
        initial_direction=dict(
            argstr='-initdirection %s',
            units='voxels',
        ),
        inputmodel=dict(
            argstr='%s',
            position=-3,
            usedefault=True,
        ),
        mask_file=dict(
            argstr='-mask %s',
            extensions=None,
            xor=['mask_file', 'mask_spec'],
        ),
        mask_spec=dict(
            argstr='-mask %s',
            position=2,
            sep=',',
            units='mm',
            xor=['mask_file', 'mask_spec'],
        ),
        maximum_number_of_tracks=dict(argstr='-maxnum %d', ),
        maximum_tract_length=dict(
            argstr='-length %s',
            units='mm',
        ),
        minimum_radius_of_curvature=dict(
            argstr='-curvature %s',
            units='mm',
        ),
        minimum_tract_length=dict(
            argstr='-minlength %s',
            units='mm',
        ),
        no_mask_interpolation=dict(argstr='-nomaskinterp', ),
        out_file=dict(
            argstr='%s',
            extensions=None,
            name_source=['in_file'],
            name_template='%s_tracked.tck',
            output_name='tracked',
            position=-1,
        ),
        seed_file=dict(
            argstr='-seed %s',
            extensions=None,
            xor=['seed_file', 'seed_spec'],
        ),
        seed_spec=dict(
            argstr='-seed %s',
            position=2,
            sep=',',
            units='mm',
            xor=['seed_file', 'seed_spec'],
        ),
        step_size=dict(
            argstr='-step %s',
            units='mm',
        ),
        stop=dict(argstr='-stop', ),
        unidirectional=dict(argstr='-unidirectional', ),
    )
    inputs = DiffusionTensorStreamlineTrack.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_DiffusionTensorStreamlineTrack_outputs():
    output_map = dict(tracked=dict(extensions=None, ), )
    outputs = DiffusionTensorStreamlineTrack.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import division

from copy import deepcopy
import os

from nibabel import Nifti1Image
import numpy as np

import pytest
import numpy.testing as npt
from nipype.testing import example_data
from nipype.interfaces.base import Bunch, TraitError
from nipype.algorithms.modelgen import (bids_gen_info, SpecifyModel,
                                        SpecifySparseModel, SpecifySPMModel)


def test_bids_gen_info():
    fname = example_data('events.tsv')
    res = bids_gen_info([fname])
    assert res[0].onsets == [[183.75, 313.75, 483.75, 633.75, 783.75, 933.75, 1083.75, 1233.75]]
    assert res[0].durations == [[20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0, 20.0]]
    assert res[0].amplitudes ==[[1, 1, 1, 1, 1, 1, 1, 1]]
    assert res[0].conditions == ['ev0']


def test_modelgen1(tmpdir):
    filename1 = tmpdir.join('test1.nii').strpath
    filename2 = tmpdir.join('test2.nii').strpath
    Nifti1Image(np.random.rand(10, 10, 10, 200),
                np.eye(4)).to_filename(filename1)
    Nifti1Image(np.random.rand(10, 10, 10, 200),
                np.eye(4)).to_filename(filename2)
    s = SpecifyModel()
    s.inputs.input_units = 'scans'
    set_output_units = lambda: setattr(s.inputs, 'output_units', 'scans')
    with pytest.raises(TraitError):
        set_output_units()
    s.inputs.functional_runs = [filename1, filename2]
    s.inputs.time_repetition = 6
    s.inputs.high_pass_filter_cutoff = 128.
    info = [
        Bunch(
            conditions=['cond1'],
            onsets=[[2, 50, 100, 180]],
            durations=[[1]],
            amplitudes=None,
            pmod=None,
            regressors=None,
            regressor_names=None,
            tmod=None),
        Bunch(
            conditions=['cond1'],
            onsets=[[30, 40, 100, 150]],
            durations=[[1]],
            amplitudes=None,
            pmod=None,
            regressors=None,
            regressor_names=None,
            tmod=None)
    ]
    s.inputs.subject_info = info
    res = s.run()
    assert len(res.outputs.session_info) == 2
    assert len(res.outputs.session_info[0]['regress']) == 0
    assert len(res.outputs.session_info[0]['cond']) == 1
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['onset']),
        np.array([12, 300, 600, 1080]))
    info = [
        Bunch(conditions=['cond1'], onsets=[[2]], durations=[[1]]),
        Bunch(conditions=['cond1'], onsets=[[3]], durations=[[1]])
    ]
    s.inputs.subject_info = deepcopy(info)
    res = s.run()
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['duration']),
        np.array([6.]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[1]['cond'][0]['duration']),
        np.array([6.]))
    info = [
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2]],
            durations=[[1, 1], [1]]),
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2, 4]],
            durations=[[1, 1], [1, 1]])
    ]
    s.inputs.subject_info = deepcopy(info)
    s.inputs.input_units = 'scans'
    res = s.run()
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['duration']),
        np.array([6., 6.]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][1]['duration']),
        np.array([
            6.,
        ]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[1]['cond'][1]['duration']),
        np.array([6., 6.]))


def test_modelgen_spm_concat(tmpdir):
    filename1 = tmpdir.join('test1.nii').strpath
    filename2 = tmpdir.join('test2.nii').strpath
    Nifti1Image(np.random.rand(10, 10, 10, 30),
                np.eye(4)).to_filename(filename1)
    Nifti1Image(np.random.rand(10, 10, 10, 30),
                np.eye(4)).to_filename(filename2)

    # Test case when only one duration is passed, as being the same for all onsets.
    s = SpecifySPMModel()
    s.inputs.input_units = 'secs'
    s.inputs.concatenate_runs = True
    setattr(s.inputs, 'output_units', 'secs')
    assert s.inputs.output_units == 'secs'
    s.inputs.functional_runs = [filename1, filename2]
    s.inputs.time_repetition = 6
    s.inputs.high_pass_filter_cutoff = 128.
    info = [
        Bunch(
            conditions=['cond1'], onsets=[[2, 50, 100, 170]], durations=[[1]]),
        Bunch(
            conditions=['cond1'], onsets=[[30, 40, 100, 150]], durations=[[1]])
    ]
    s.inputs.subject_info = deepcopy(info)
    res = s.run()
    assert len(res.outputs.session_info) == 1
    assert len(res.outputs.session_info[0]['regress']) == 1
    assert np.sum(res.outputs.session_info[0]['regress'][0]['val']) == 30
    assert len(res.outputs.session_info[0]['cond']) == 1
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['onset']),
        np.array([2.0, 50.0, 100.0, 170.0, 210.0, 220.0, 280.0, 330.0]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['duration']),
        np.array([1., 1., 1., 1., 1., 1., 1., 1.]))

    # Test case of scans as output units instead of seconds
    setattr(s.inputs, 'output_units', 'scans')
    assert s.inputs.output_units == 'scans'
    s.inputs.subject_info = deepcopy(info)
    res = s.run()
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['onset']),
        np.array([2.0, 50.0, 100.0, 170.0, 210.0, 220.0, 280.0, 330.0]) / 6)

    # Test case for no concatenation with seconds as output units
    s.inputs.concatenate_runs = False
    s.inputs.subject_info = deepcopy(info)
    s.inputs.output_units = 'secs'
    res = s.run()
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['onset']),
        np.array([2.0, 50.0, 100.0, 170.0]))

    # Test case for variable number of events in separate runs, sometimes unique.
    filename3 = tmpdir.join('test3.nii').strpath
    Nifti1Image(np.random.rand(10, 10, 10, 30),
                np.eye(4)).to_filename(filename3)
    s.inputs.functional_runs = [filename1, filename2, filename3]
    info = [
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2]],
            durations=[[1, 1], [1]]),
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2, 4]],
            durations=[[1, 1], [1, 1]]),
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2]],
            durations=[[1, 1], [1]])
    ]
    s.inputs.subject_info = deepcopy(info)
    res = s.run()
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['duration']),
        np.array([1., 1.]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][1]['duration']),
        np.array([
            1.,
        ]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[1]['cond'][1]['duration']),
        np.array([1., 1.]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[2]['cond'][1]['duration']),
        np.array([
            1.,
        ]))

    # Test case for variable number of events in concatenated runs, sometimes unique.
    s.inputs.concatenate_runs = True
    info = [
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2]],
            durations=[[1, 1], [1]]),
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2, 4]],
            durations=[[1, 1], [1, 1]]),
        Bunch(
            conditions=['cond1', 'cond2'],
            onsets=[[2, 3], [2]],
            durations=[[1, 1], [1]])
    ]
    s.inputs.subject_info = deepcopy(info)
    res = s.run()
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][0]['duration']),
        np.array([1., 1., 1., 1., 1., 1.]))
    npt.assert_almost_equal(
        np.array(res.outputs.session_info[0]['cond'][1]['duration']),
        np.array([1., 1., 1., 1.]))


def test_modelgen_sparse(tmpdir):
    filename1 = tmpdir.join('test1.nii').strpath
    filename2 = tmpdir.join('test2.nii').strpath
    Nifti1Image(np.random.rand(10, 10, 10, 50),
                np.eye(4)).to_filename(filename1)
    Nifti1Image(np.random.rand(10, 10, 10, 50),
                np.eye(4)).to_filename(filename2)
    s = SpecifySparseModel()
    s.inputs.input_units = 'secs'
    s.inputs.functional_runs = [filename1, filename2]
    s.inputs.time_repetition = 6
    info = [
        Bunch(
            conditions=['cond1'], onsets=[[0, 50, 100, 180]], durations=[[2]]),
        Bunch(
            conditions=['cond1'], onsets=[[30, 40, 100, 150]], durations=[[1]])
    ]
    s.inputs.subject_info = info
    s.inputs.volumes_in_cluster = 1
    s.inputs.time_acquisition = 2
    s.inputs.high_pass_filter_cutoff = np.inf
    res = s.run()
    assert len(res.outputs.session_info) == 2
    assert len(res.outputs.session_info[0]['regress']) == 1
    assert len(res.outputs.session_info[0]['cond']) == 0

    s.inputs.stimuli_as_impulses = False
    res = s.run()
    assert res.outputs.session_info[0]['regress'][0]['val'][0] == 1.0

    s.inputs.model_hrf = True
    res = s.run()
    npt.assert_almost_equal(
        res.outputs.session_info[0]['regress'][0]['val'][0],
        0.016675298129743384)
    assert len(res.outputs.session_info[0]['regress']) == 1
    s.inputs.use_temporal_deriv = True
    res = s.run()

    assert len(res.outputs.session_info[0]['regress']) == 2
    npt.assert_almost_equal(
        res.outputs.session_info[0]['regress'][0]['val'][0],
        0.016675298129743384)
    npt.assert_almost_equal(
        res.outputs.session_info[1]['regress'][1]['val'][5],
        0.007671459162258378)



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..brains import BRAINSMush


def test_BRAINSMush_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        boundingBoxSize=dict(
            argstr='--boundingBoxSize %s',
            sep=',',
        ),
        boundingBoxStart=dict(
            argstr='--boundingBoxStart %s',
            sep=',',
        ),
        desiredMean=dict(argstr='--desiredMean %f', ),
        desiredVariance=dict(argstr='--desiredVariance %f', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputFirstVolume=dict(
            argstr='--inputFirstVolume %s',
            extensions=None,
        ),
        inputMaskVolume=dict(
            argstr='--inputMaskVolume %s',
            extensions=None,
        ),
        inputSecondVolume=dict(
            argstr='--inputSecondVolume %s',
            extensions=None,
        ),
        lowerThresholdFactor=dict(argstr='--lowerThresholdFactor %f', ),
        lowerThresholdFactorPre=dict(argstr='--lowerThresholdFactorPre %f', ),
        numberOfThreads=dict(argstr='--numberOfThreads %d', ),
        outputMask=dict(
            argstr='--outputMask %s',
            hash_files=False,
        ),
        outputVolume=dict(
            argstr='--outputVolume %s',
            hash_files=False,
        ),
        outputWeightsFile=dict(
            argstr='--outputWeightsFile %s',
            hash_files=False,
        ),
        seed=dict(
            argstr='--seed %s',
            sep=',',
        ),
        upperThresholdFactor=dict(argstr='--upperThresholdFactor %f', ),
        upperThresholdFactorPre=dict(argstr='--upperThresholdFactorPre %f', ),
    )
    inputs = BRAINSMush.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_BRAINSMush_outputs():
    output_map = dict(
        outputMask=dict(extensions=None, ),
        outputVolume=dict(extensions=None, ),
        outputWeightsFile=dict(extensions=None, ),
    )
    outputs = BRAINSMush.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import ApplyVolTransform


def test_ApplyVolTransform_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fs_target=dict(
            argstr='--fstarg',
            mandatory=True,
            requires=['reg_file'],
            xor=('target_file', 'tal', 'fs_target'),
        ),
        fsl_reg_file=dict(
            argstr='--fsl %s',
            extensions=None,
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
        interp=dict(argstr='--interp %s', ),
        inverse=dict(argstr='--inv', ),
        invert_morph=dict(
            argstr='--inv-morph',
            requires=['m3z_file'],
        ),
        lta_file=dict(
            argstr='--lta %s',
            extensions=None,
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
        lta_inv_file=dict(
            argstr='--lta-inv %s',
            extensions=None,
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
        m3z_file=dict(
            argstr='--m3z %s',
            extensions=None,
        ),
        mni_152_reg=dict(
            argstr='--regheader',
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
        no_ded_m3z_path=dict(
            argstr='--noDefM3zPath',
            requires=['m3z_file'],
        ),
        no_resample=dict(argstr='--no-resample', ),
        reg_file=dict(
            argstr='--reg %s',
            extensions=None,
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
        reg_header=dict(
            argstr='--regheader',
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
        source_file=dict(
            argstr='--mov %s',
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        subject=dict(
            argstr='--s %s',
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
        subjects_dir=dict(),
        tal=dict(
            argstr='--tal',
            mandatory=True,
            xor=('target_file', 'tal', 'fs_target'),
        ),
        tal_resolution=dict(argstr='--talres %.10f', ),
        target_file=dict(
            argstr='--targ %s',
            extensions=None,
            mandatory=True,
            xor=('target_file', 'tal', 'fs_target'),
        ),
        transformed_file=dict(
            argstr='--o %s',
            extensions=None,
            genfile=True,
        ),
        xfm_reg_file=dict(
            argstr='--xfm %s',
            extensions=None,
            mandatory=True,
            xor=('reg_file', 'lta_file', 'lta_inv_file', 'fsl_reg_file',
                 'xfm_reg_file', 'reg_header', 'mni_152_reg', 'subject'),
        ),
    )
    inputs = ApplyVolTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_ApplyVolTransform_outputs():
    output_map = dict(transformed_file=dict(extensions=None, ), )
    outputs = ApplyVolTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from builtins import open, range
import os

import nipype.interfaces.fsl.dti as fsl
from nipype.interfaces.fsl import Info, no_fsl
from nipype.interfaces.base import Undefined

import pytest
from nipype.testing.fixtures import create_files_in_directory


# test dtifit
@pytest.mark.skipif(no_fsl(), reason="fsl is not installed")
def test_dtifit2(create_files_in_directory):
    filelist, outdir = create_files_in_directory
    dti = fsl.DTIFit()
    # make sure command gets called
    assert dti.cmd == 'dtifit'

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        dti.run()

    # .inputs based parameters setting
    dti.inputs.dwi = filelist[0]
    dti.inputs.base_name = 'foo.dti.nii'
    dti.inputs.mask = filelist[1]
    dti.inputs.bvecs = filelist[0]
    dti.inputs.bvals = filelist[1]
    dti.inputs.min_z = 10
    dti.inputs.max_z = 50

    assert dti.cmdline == \
        'dtifit -k %s -o foo.dti.nii -m %s -r %s -b %s -Z 50 -z 10' % (filelist[0],
                                                                       filelist[1],
                                                                       filelist[0],
                                                                       filelist[1])


@pytest.mark.xfail(
    reason="These tests are skipped until we clean up some of this code")
def test_randomise2():

    rand = fsl.Randomise()

    # make sure command gets called
    assert rand.cmd == 'randomise'

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        rand.run()

    # .inputs based parameters setting
    rand.inputs.input_4D = 'infile.nii'
    rand.inputs.output_rootname = 'outfile'
    rand.inputs.design_matrix = 'design.mat'
    rand.inputs.t_contrast = 'infile.con'

    actualCmdline = sorted(rand.cmdline.split())
    cmd = 'randomise -i infile.nii -o outfile -d design.mat -t infile.con'
    desiredCmdline = sorted(cmd.split())
    assert actualCmdline == desiredCmdline

    # .run based parameter setting
    rand2 = fsl.Randomise(
        input_4D='infile2',
        output_rootname='outfile2',
        f_contrast='infile.f',
        one_sample_gmean=True,
        int_seed=4)

    actualCmdline = sorted(rand2.cmdline.split())
    cmd = 'randomise -i infile2 -o outfile2 -1 -f infile.f --seed=4'
    desiredCmdline = sorted(cmd.split())
    assert actualCmdline == desiredCmdline

    rand3 = fsl.Randomise()
    results = rand3.run(input_4D='infile3', output_rootname='outfile3')
    assert results.runtime.cmdline == \
        'randomise -i infile3 -o outfile3'

    # test arguments for opt_map
    opt_map = {
        'demean_data': ('-D', True),
        'one_sample_gmean': ('-1', True),
        'mask_image': ('-m inp_mask', 'inp_mask'),
        'design_matrix': ('-d design.mat', 'design.mat'),
        't_contrast': ('-t input.con', 'input.con'),
        'f_contrast': ('-f input.fts', 'input.fts'),
        'xchange_block_labels': ('-e design.grp', 'design.grp'),
        'print_unique_perm': ('-q', True),
        'print_info_parallelMode': ('-Q', True),
        'num_permutations': ('-n 10', 10),
        'vox_pvalus': ('-x', True),
        'fstats_only': ('--fonly', True),
        'thresh_free_cluster': ('-T', True),
        'thresh_free_cluster_2Dopt': ('--T2', True),
        'cluster_thresholding': ('-c 0.20', 0.20),
        'cluster_mass_thresholding': ('-C 0.40', 0.40),
        'fcluster_thresholding': ('-F 0.10', 0.10),
        'fcluster_mass_thresholding': ('-S 0.30', 0.30),
        'variance_smoothing': ('-v 0.20', 0.20),
        'diagnostics_off': ('--quiet', True),
        'output_raw': ('-R', True),
        'output_perm_vect': ('-P', True),
        'int_seed': ('--seed=20', 20),
        'TFCE_height_param': ('--tfce_H=0.11', 0.11),
        'TFCE_extent_param': ('--tfce_E=0.50', 0.50),
        'TFCE_connectivity': ('--tfce_C=0.30', 0.30),
        'list_num_voxel_EVs_pos': ('--vxl=1,2,3,4', '1,2,3,4'),
        'list_img_voxel_EVs': ('--vxf=6,7,8,9,3', '6,7,8,9,3')
    }

    for name, settings in list(opt_map.items()):
        rand4 = fsl.Randomise(
            input_4D='infile', output_rootname='root', **{
                name: settings[1]
            })
        assert rand4.cmdline == rand4.cmd + ' -i infile -o root ' + settings[0]


@pytest.mark.xfail(
    reason="These tests are skipped until we clean up some of this code")
def test_Randomise_parallel():
    rand = fsl.Randomise_parallel()

    # make sure command gets called
    assert rand.cmd == 'randomise_parallel'

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        rand.run()

    # .inputs based parameters setting
    rand.inputs.input_4D = 'infile.nii'
    rand.inputs.output_rootname = 'outfile'
    rand.inputs.design_matrix = 'design.mat'
    rand.inputs.t_contrast = 'infile.con'

    actualCmdline = sorted(rand.cmdline.split())
    cmd = ('randomise_parallel -i infile.nii -o outfile -d design.mat -t '
           'infile.con')
    desiredCmdline = sorted(cmd.split())
    assert actualCmdline == desiredCmdline

    # .run based parameter setting
    rand2 = fsl.Randomise_parallel(
        input_4D='infile2',
        output_rootname='outfile2',
        f_contrast='infile.f',
        one_sample_gmean=True,
        int_seed=4)

    actualCmdline = sorted(rand2.cmdline.split())
    cmd = 'randomise_parallel -i infile2 -o outfile2 -1 -f infile.f --seed=4'
    desiredCmdline = sorted(cmd.split())
    assert actualCmdline == desiredCmdline

    rand3 = fsl.Randomise_parallel()
    results = rand3.run(input_4D='infile3', output_rootname='outfile3')
    assert results.runtime.cmdline == \
        'randomise_parallel -i infile3 -o outfile3'

    # test arguments for opt_map
    opt_map = {
        'demean_data': ('-D', True),
        'one_sample_gmean': ('-1', True),
        'mask_image': ('-m inp_mask', 'inp_mask'),
        'design_matrix': ('-d design.mat', 'design.mat'),
        't_contrast': ('-t input.con', 'input.con'),
        'f_contrast': ('-f input.fts', 'input.fts'),
        'xchange_block_labels': ('-e design.grp', 'design.grp'),
        'print_unique_perm': ('-q', True),
        'print_info_parallelMode': ('-Q', True),
        'num_permutations': ('-n 10', 10),
        'vox_pvalus': ('-x', True),
        'fstats_only': ('--fonly', True),
        'thresh_free_cluster': ('-T', True),
        'thresh_free_cluster_2Dopt': ('--T2', True),
        'cluster_thresholding': ('-c 0.20', 0.20),
        'cluster_mass_thresholding': ('-C 0.40', 0.40),
        'fcluster_thresholding': ('-F 0.10', 0.10),
        'fcluster_mass_thresholding': ('-S 0.30', 0.30),
        'variance_smoothing': ('-v 0.20', 0.20),
        'diagnostics_off': ('--quiet', True),
        'output_raw': ('-R', True),
        'output_perm_vect': ('-P', True),
        'int_seed': ('--seed=20', 20),
        'TFCE_height_param': ('--tfce_H=0.11', 0.11),
        'TFCE_extent_param': ('--tfce_E=0.50', 0.50),
        'TFCE_connectivity': ('--tfce_C=0.30', 0.30),
        'list_num_voxel_EVs_pos': ('--vxl=' + repr([1, 2, 3, 4]),
                                   repr([1, 2, 3, 4])),
        'list_img_voxel_EVs': ('--vxf=' + repr([6, 7, 8, 9, 3]),
                               repr([6, 7, 8, 9, 3]))
    }

    for name, settings in list(opt_map.items()):
        rand4 = fsl.Randomise_parallel(
            input_4D='infile', output_rootname='root', **{
                name: settings[1]
            })
        assert rand4.cmdline == rand4.cmd + ' -i infile -o root ' + settings[0]


# test proj_thresh
@pytest.mark.xfail(
    reason="These tests are skipped until we clean up some of this code")
def test_Proj_thresh():
    proj = fsl.ProjThresh()

    # make sure command gets called
    assert proj.cmd == 'proj_thresh'

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        proj.run()

    # .inputs based parameters setting
    proj.inputs.volumes = ['vol1', 'vol2', 'vol3']
    proj.inputs.threshold = 3
    assert proj.cmdline == 'proj_thresh vol1 vol2 vol3 3'

    proj2 = fsl.ProjThresh(threshold=10, volumes=['vola', 'volb'])
    assert proj2.cmdline == 'proj_thresh vola volb 10'

    # .run based parameters setting
    proj3 = fsl.ProjThresh()
    results = proj3.run(volumes=['inp1', 'inp3', 'inp2'], threshold=2)
    assert results.runtime.cmdline == 'proj_thresh inp1 inp3 inp2 2'
    assert results.runtime.returncode != 0
    assert isinstance(results.interface.inputs.volumes, list)
    assert results.interface.inputs.threshold == 2

    # test arguments for opt_map
    # Proj_thresh doesn't have an opt_map{}


# test vec_reg
@pytest.mark.xfail(
    reason="These tests are skipped until we clean up some of this code")
def test_Vec_reg():

    vrg = fsl.VecReg()

    # make sure command gets called
    assert vrg.cmd == 'vecreg'

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        vrg.run()

    # .inputs based parameters setting
    vrg.inputs.infile = 'infile'
    vrg.inputs.outfile = 'outfile'
    vrg.inputs.refVolName = 'MNI152'
    vrg.inputs.affineTmat = 'tmat.mat'
    assert vrg.cmdline == 'vecreg -i infile -o outfile -r MNI152 -t tmat.mat'

    # .run based parameter setting
    vrg2 = fsl.VecReg(
        infile='infile2',
        outfile='outfile2',
        refVolName='MNI152',
        affineTmat='tmat2.mat',
        brainMask='nodif_brain_mask')

    actualCmdline = sorted(vrg2.cmdline.split())
    cmd = 'vecreg -i infile2 -o outfile2 -r MNI152 -t tmat2.mat -m nodif_brain_mask'
    desiredCmdline = sorted(cmd.split())
    assert actualCmdline == desiredCmdline

    vrg3 = fsl.VecReg()
    results = vrg3.run(
        infile='infile3',
        outfile='outfile3',
        refVolName='MNI152',
        affineTmat='tmat3.mat',
    )

    assert results.runtime.cmdline == \
        'vecreg -i infile3 -o outfile3 -r MNI152 -t tmat3.mat'
    assert results.runtime.returncode != 0
    assert results.interface.inputs.infile == 'infile3'
    assert results.interface.inputs.outfile == 'outfile3'
    assert results.interface.inputs.refVolName == 'MNI152'
    assert results.interface.inputs.affineTmat == 'tmat3.mat'

    # test arguments for opt_map
    opt_map = {
        'verbose': ('-v', True),
        'helpDoc': ('-h', True),
        'tensor': ('--tensor', True),
        'affineTmat': ('-t Tmat', 'Tmat'),
        'warpFile': ('-w wrpFile', 'wrpFile'),
        'interpolation': ('--interp=sinc', 'sinc'),
        'brainMask': ('-m mask', 'mask')
    }

    for name, settings in list(opt_map.items()):
        vrg4 = fsl.VecReg(
            infile='infile',
            outfile='outfile',
            refVolName='MNI152',
            **{
                name: settings[1]
            })
        assert vrg4.cmdline == vrg4.cmd + \
            ' -i infile -o outfile -r MNI152 ' + settings[0]


# test find_the_biggest
@pytest.mark.xfail(
    reason="These tests are skipped until we clean up some of this code")
def test_Find_the_biggest():
    fbg = fsl.FindTheBiggest()

    # make sure command gets called
    assert fbg.cmd == 'find_the_biggest'

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        fbg.run()

    # .inputs based parameters setting
    fbg.inputs.infiles = 'seed*'
    fbg.inputs.outfile = 'fbgfile'
    assert fbg.cmdline == 'find_the_biggest seed* fbgfile'

    fbg2 = fsl.FindTheBiggest(infiles='seed2*', outfile='fbgfile2')
    assert fbg2.cmdline == 'find_the_biggest seed2* fbgfile2'

    # .run based parameters setting
    fbg3 = fsl.FindTheBiggest()
    results = fbg3.run(infiles='seed3', outfile='out3')
    assert results.runtime.cmdline == 'find_the_biggest seed3 out3'

    # test arguments for opt_map
    # Find_the_biggest doesn't have an opt_map{}


@pytest.mark.skipif(no_fsl(), reason="fsl is not installed")
def test_tbss_skeleton(create_files_in_directory):
    skeletor = fsl.TractSkeleton()

    files, newdir = create_files_in_directory

    # Test the underlying command
    assert skeletor.cmd == "tbss_skeleton"

    # It shouldn't run yet
    with pytest.raises(ValueError):
        skeletor.run()

    # Test the most basic way to use it
    skeletor.inputs.in_file = files[0]

    # First by implicit argument
    skeletor.inputs.skeleton_file = True
    assert skeletor.cmdline == \
        "tbss_skeleton -i a.nii -o %s" % os.path.join(newdir, "a_skeleton.nii")

    # Now with a specific name
    skeletor.inputs.skeleton_file = "old_boney.nii"
    assert skeletor.cmdline == "tbss_skeleton -i a.nii -o old_boney.nii"

    # Now test the more complicated usage
    bones = fsl.TractSkeleton(in_file="a.nii", project_data=True)

    # This should error
    with pytest.raises(ValueError):
        bones.run()

    # But we can set what we need
    bones.inputs.threshold = 0.2
    bones.inputs.distance_map = "b.nii"
    bones.inputs.data_file = "b.nii"  # Even though that's silly

    # Now we get a command line
    assert bones.cmdline == \
        "tbss_skeleton -i a.nii -p 0.200 b.nii %s b.nii %s" % (Info.standard_image("LowerCingulum_1mm.nii.gz"),
                                                               os.path.join(newdir, "b_skeletonised.nii"))

    # Can we specify a mask?
    bones.inputs.use_cingulum_mask = Undefined
    bones.inputs.search_mask_file = "a.nii"
    assert bones.cmdline == \
        "tbss_skeleton -i a.nii -p 0.200 b.nii a.nii b.nii %s" % os.path.join(newdir, "b_skeletonised.nii")


@pytest.mark.skipif(no_fsl(), reason="fsl is not installed")
def test_distancemap(create_files_in_directory):
    mapper = fsl.DistanceMap()

    files, newdir = create_files_in_directory

    # Test the underlying command
    assert mapper.cmd == "distancemap"

    # It shouldn't run yet
    with pytest.raises(ValueError):
        mapper.run()

    # But if we do this...
    mapper.inputs.in_file = "a.nii"

    # It should
    assert mapper.cmdline == "distancemap --out=%s --in=a.nii" % os.path.join(
        newdir, "a_dstmap.nii")

    # And we should be able to write out a maxima map
    mapper.inputs.local_max_file = True
    assert mapper.cmdline == \
        "distancemap --out=%s --in=a.nii --localmax=%s" % (os.path.join(newdir, "a_dstmap.nii"),
                                                           os.path.join(newdir, "a_lclmax.nii"))

    # And call it whatever we want
    mapper.inputs.local_max_file = "max.nii"
    assert mapper.cmdline == \
        "distancemap --out=%s --in=a.nii --localmax=max.nii" % os.path.join(newdir, "a_dstmap.nii")



# -*- coding: utf-8 -*-
# coding: utf-8
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Generic interfaces to manipulate registration parameters files, including
transform files (to configure warpings)

"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
from builtins import open

import os.path as op

from ... import logging
from ..base import (BaseInterface, BaseInterfaceInputSpec, isdefined,
                    TraitedSpec, File, traits)
iflogger = logging.getLogger('nipype.interface')


class EditTransformInputSpec(BaseInterfaceInputSpec):
    transform_file = File(
        exists=True, mandatory=True, desc='transform-parameter file, only 1')
    reference_image = File(
        exists=True,
        desc=('set a new reference image to change the '
              'target coordinate system.'))
    interpolation = traits.Enum(
        'cubic',
        'linear',
        'nearest',
        usedefault=True,
        argstr='FinalBSplineInterpolationOrder',
        desc='set a new interpolator for transformation')

    output_type = traits.Enum(
        'float',
        'unsigned char',
        'unsigned short',
        'short',
        'unsigned long',
        'long',
        'double',
        argstr='ResultImagePixelType',
        desc='set a new output pixel type for resampled images')
    output_format = traits.Enum(
        'nii.gz',
        'nii',
        'mhd',
        'hdr',
        'vtk',
        argstr='ResultImageFormat',
        desc='set a new image format for resampled images')
    output_file = File(desc='the filename for the resulting transform file')


class EditTransformOutputSpec(TraitedSpec):
    output_file = File(exists=True, desc='output transform file')


class EditTransform(BaseInterface):
    """
    Manipulates an existing transform file generated with elastix

    Example
    -------

    >>> from nipype.interfaces.elastix import EditTransform
    >>> tfm = EditTransform()
    >>> tfm.inputs.transform_file = 'TransformParameters.0.txt'  # doctest: +SKIP
    >>> tfm.inputs.reference_image = 'fixed1.nii'  # doctest: +SKIP
    >>> tfm.inputs.output_type = 'unsigned char'
    >>> tfm.run() # doctest: +SKIP

    """

    input_spec = EditTransformInputSpec
    output_spec = EditTransformOutputSpec
    _out_file = ''
    _pattern = '\((?P<entry>%s\s\"?)([-\.\s\w]+)(\"?\))'

    _interp = {'nearest': 0, 'linear': 1, 'cubic': 3}

    def _run_interface(self, runtime):
        import re
        import nibabel as nb
        import numpy as np

        contents = ''

        with open(self.inputs.transform_file, 'r') as f:
            contents = f.read()

        if isdefined(self.inputs.output_type):
            p = re.compile((self._pattern %
                            'ResultImagePixelType').decode('string-escape'))
            rep = '(\g<entry>%s\g<3>' % self.inputs.output_type
            contents = p.sub(rep, contents)

        if isdefined(self.inputs.output_format):
            p = re.compile(
                (self._pattern % 'ResultImageFormat').decode('string-escape'))
            rep = '(\g<entry>%s\g<3>' % self.inputs.output_format
            contents = p.sub(rep, contents)

        if isdefined(self.inputs.interpolation):
            p = re.compile(
                (self._pattern %
                 'FinalBSplineInterpolationOrder').decode('string-escape'))
            rep = '(\g<entry>%s\g<3>' % self._interp[self.inputs.interpolation]
            contents = p.sub(rep, contents)

        if isdefined(self.inputs.reference_image):
            im = nb.load(self.inputs.reference_image)

            if len(im.header.get_zooms()) == 4:
                im = nb.func.four_to_three(im)[0]

            size = ' '.join(["%01d" % s for s in im.shape])
            p = re.compile((self._pattern % 'Size').decode('string-escape'))
            rep = '(\g<entry>%s\g<3>' % size
            contents = p.sub(rep, contents)

            index = ' '.join(["0" for s in im.shape])
            p = re.compile((self._pattern % 'Index').decode('string-escape'))
            rep = '(\g<entry>%s\g<3>' % index
            contents = p.sub(rep, contents)

            spacing = ' '.join(["%0.4f" % f for f in im.header.get_zooms()])
            p = re.compile((self._pattern % 'Spacing').decode('string-escape'))
            rep = '(\g<entry>%s\g<3>' % spacing
            contents = p.sub(rep, contents)

            itkmat = np.eye(4)
            itkmat[0, 0] = -1
            itkmat[1, 1] = -1

            affine = np.dot(itkmat, im.affine)
            dirs = ' '.join(
                ['%0.4f' % f for f in affine[0:3, 0:3].reshape(-1)])
            orig = ' '.join(['%0.4f' % f for f in affine[0:3, 3].reshape(-1)])

            # p = re.compile((self._pattern % 'Direction').decode('string-escape'))
            # rep = '(\g<entry>%s\g<3>' % dirs
            # contents = p.sub(rep, contents)

            p = re.compile((self._pattern % 'Origin').decode('string-escape'))
            rep = '(\g<entry>%s\g<3>' % orig
            contents = p.sub(rep, contents)

        with open(self._get_outfile(), 'w') as of:
            of.write(contents)

        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['output_file'] = getattr(self, '_out_file')
        return outputs

    def _get_outfile(self):
        val = getattr(self, '_out_file')
        if val is not None and val != '':
            return val

        if isdefined(self.inputs.output_file):
            setattr(self, '_out_file', self.inputs.output_file)
            return self.inputs.output_file

        out_file = op.abspath(op.basename(self.inputs.transform_file))
        setattr(self, '_out_file', out_file)
        return out_file



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)
import os

from ..base import (traits, TraitedSpec, File, CommandLine,
                    CommandLineInputSpec, InputMultiPath)
from ...utils.filemanip import split_filename


class ImageStatsInputSpec(CommandLineInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        argstr='-images %s',
        mandatory=True,
        position=-1,
        desc=('List of images to process. They must '
              'be in the same space and have the same '
              'dimensions.'))
    stat = traits.Enum(
        "min",
        "max",
        "mean",
        "median",
        "sum",
        "std",
        "var",
        argstr='-stat %s',
        units='NA',
        mandatory=True,
        desc="The statistic to compute.")

    out_type = traits.Enum(
        "float",
        "char",
        "short",
        "int",
        "long",
        "double",
        argstr='-outputdatatype %s',
        usedefault=True,
        desc=('A Camino data type string, default is "float". '
              'Type must be signed.'))
    output_root = File(
        argstr='-outputroot %s',
        mandatory=True,
        desc=('Filename root prepended onto the names of the output '
              ' files. The extension will be determined from the input.'))


class ImageStatsOutputSpec(TraitedSpec):
    out_file = File(
        exists=True,
        desc='Path of the file computed with the statistic chosen')


class ImageStats(CommandLine):
    """
    This program computes voxelwise statistics on a series of 3D images. The images
    must be in the same space; the operation is performed voxelwise and one output
    is produced per voxel.

    Examples
    --------

    >>> import nipype.interfaces.camino as cam
    >>> imstats = cam.ImageStats()
    >>> imstats.inputs.in_files = ['im1.nii','im2.nii','im3.nii']
    >>> imstats.inputs.stat = 'max'
    >>> imstats.run()                  # doctest: +SKIP
    """
    _cmd = 'imagestats'
    input_spec = ImageStatsInputSpec
    output_spec = ImageStatsOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        output_root = self.inputs.output_root
        first_file = self.inputs.in_files[0]
        _, _, ext = split_filename(first_file)
        return output_root + ext
